# coding: utf-8

"""
Collections that wrap multiple targets.
"""

__all__ = [
    "TargetCollection", "FileCollection", "SiblingFileCollection", "NestedSiblingFileCollection",
]


import types
import random
from functools import partial
from contextlib import contextmanager
from collections import defaultdict, deque

import six

from law.config import Config
from law.target.base import Target
from law.target.file import FileSystemTarget, FileSystemDirectoryTarget, localize_file_targets
from law.target.mirrored import MirroredTarget, MirroredDirectoryTarget
from law.target.local import LocalDirectoryTarget
from law.util import no_value, colored, flatten, map_struct
from law.logger import get_logger


logger = get_logger(__name__)


class TargetCollection(Target):
    """
    Collection of arbitrary targets.
    """

    def __init__(self, targets, threshold=1.0, optional_existing=None, **kwargs):
        if isinstance(targets, types.GeneratorType):
            targets = list(targets)
        elif not isinstance(targets, (list, tuple, dict)):
            raise TypeError("invalid targets, must be of type: list, tuple, dict")

        super(TargetCollection, self).__init__(**kwargs)

        # store attributes
        self.targets = targets
        self.threshold = threshold
        self.optional_existing = optional_existing

        # store flat targets per element in the input structure of targets
        if isinstance(targets, (list, tuple)):
            gen = (flatten(t) for t in targets)
        else:  # dict
            gen = ((k, flatten(t)) for k, t in six.iteritems(targets))
        self._flat_targets = targets.__class__(gen)

        # also store an entirely flat list of targets for simplified iterations
        self._flat_target_list = flatten(targets)

    def __len__(self):
        return len(self.targets)

    def __getitem__(self, key):
        return self.targets[key]

    def __iter__(self):
        # explicitly disable iterability enabled by __getitem__ as per PEP234
        # to (e.g.) prevent that flatten() applies to collections
        raise TypeError("'{}' object is not iterable".format(self.__class__.__name__))

    def _copy_kwargs(self):
        kwargs = super(TargetCollection, self)._copy_kwargs()
        kwargs["threshold"] = self.threshold
        kwargs["optional_existing"] = self.optional_existing
        return kwargs

    def _repr_pairs(self):
        pairs = Target._repr_pairs(self) + [("len", len(self))]

        # add non-default attributes
        if self.threshold != 1.0:
            pairs.append(("threshold", self.threshold))
        if self.optional_existing is not None:
            pairs.append(("optional_existing", self.optional_existing))

        return pairs

    def _iter_flat(self):
        # prepare the generator for looping
        if isinstance(self._flat_targets, (list, tuple)):
            gen = enumerate(self._flat_targets)
        else:  # dict
            gen = six.iteritems(self._flat_targets)

        # loop and yield
        for key, targets in gen:
            yield (key, targets)

    def _iter_state(
        self,
        existing=True,
        optional_existing=no_value,
        keys=False,
        unpack=True,
        exists_func=None,
    ):
        existing = bool(existing)
        if optional_existing is no_value:
            optional_existing = self.optional_existing

        # helper to check for existence
        if exists_func is None:
            def exists_func(t):
                if optional_existing is not None and t.optional:
                    return bool(optional_existing)
                if isinstance(t, TargetCollection):
                    return t.exists(optional_existing=optional_existing)
                return t.exists()

        # loop and yield
        for key, targets in self._iter_flat():
            state = all(map(exists_func, targets))
            if state is existing:
                if unpack:
                    targets = self.targets[key]
                yield (key, targets) if keys else targets

    def iter_existing(self, **kwargs):
        return self._iter_state(existing=True, **kwargs)

    def iter_missing(self, **kwargs):
        return self._iter_state(existing=False, **kwargs)

    def keys(self):
        if isinstance(self._flat_targets, (list, tuple)):
            return list(range(len(self)))
        # dict
        return list(self._flat_targets.keys())

    def uri(self, *args, **kwargs):
        return flatten(t.uri(*args, **kwargs) for t in self._flat_target_list)

    @property
    def first_target(self):
        if not self._flat_target_list:
            return None

        return flatten_collections(self._flat_target_list)[0]

    def remove(self, silent=True):
        for t in self._flat_target_list:
            if silent:
                t.remove(silent=True)
            elif t.exists():
                t.remove()

    def _abs_threshold(self):
        if self.threshold < 0:
            return 0
        if self.threshold <= 1:
            return len(self) * self.threshold
        return min(len(self), max(self.threshold, 0.0))

    def complete(self, **kwargs):
        kwargs["optional_existing"] = True
        return self.optional or self.exists(**kwargs)

    def _exists_fwd(self, **kwargs):
        fwd = ["optional_existing", "exists_func"]
        return self.exists(**{key: kwargs[key] for key in fwd if key in kwargs})

    def exists(self, **kwargs):
        # get the threshold
        threshold = self._abs_threshold()
        if threshold == 0:
            return True

        # simple counting with early stopping criteria for both success and fail cases
        n = 0
        for i, _ in enumerate(self.iter_existing(**kwargs)):
            n += 1

            # check for early success
            if n >= threshold:
                return True

            # check for early fail
            if n + (len(self) - i - 1) < threshold:
                return False

        return False

    def count(self, **kwargs):
        # simple counting of keys
        keys = kwargs.get("keys", False)
        kwargs["keys"] = True
        target_keys = [key for key, _ in self._iter_state(**kwargs)]

        n = len(target_keys)
        return (n, target_keys) if keys else n

    def random_target(self):
        if isinstance(self.targets, (list, tuple)):
            return random.choice(self.targets)
        # dict
        return random.choice(list(self.targets.values()))

    def map(self, func):
        """
        Returns a copy of this collection with all targets being transformed by *func*.
        """
        return self.__class__(map_struct(func, self.targets), **self._copy_kwargs())

    def status_text(self, max_depth=0, flags=None, color=False, **kwargs):
        count, existing_keys = self.count(keys=True, **kwargs)
        exists = count >= self._abs_threshold()

        if exists:
            text = "existent"
            _color = "green"
        else:
            text = "absent"
            _color = "red" if not self.optional else "dark_grey"

        text = colored(text, _color, style="bright") if color else text
        text += " ({}/{})".format(count, len(self))

        if flags and "missing" in flags and count != len(self):
            missing_keys = [str(key) for key in self.keys() if key not in existing_keys]
            text += ", missing branches: " + ",".join(missing_keys)

        if max_depth > 0:
            if isinstance(self.targets, (list, tuple)):
                gen = enumerate(self.targets)
            else:  # dict
                gen = six.iteritems(self.targets)

            for key, item in gen:
                text += "\n{}: ".format(key)

                if isinstance(item, TargetCollection):
                    t = item.status_text(max_depth=max_depth - 1, color=color, **kwargs)
                    text += "\n  ".join(t.split("\n"))
                elif isinstance(item, Target):
                    t = item.status_text(color=color, exists=key in existing_keys)
                    text += "{} ({})".format(t, item.repr(color=color))
                else:
                    t = self.__class__(item).status_text(max_depth=max_depth - 1, color=color)
                    text += "\n   ".join(t.split("\n"))

        return text


class FileCollection(TargetCollection):
    """
    Collection of targets that represent files or other FileCollection's.
    """

    def __init__(self, *args, **kwargs):
        TargetCollection.__init__(self, *args, **kwargs)

        # check if all targets are either FileSystemTarget's or FileCollection's
        for t in self._flat_target_list:
            if not isinstance(t, (FileSystemTarget, FileCollection)):
                raise TypeError("FileCollection's only wrap FileSystemTarget's and other "
                    "FileCollection's, got {}".format(t.__class__))

    @contextmanager
    def localize(self, *args, **kwargs):
        # when localizing collections using temporary files, it makes sense to put
        # them all in the same temporary directory
        tmp_dir = kwargs.get("tmp_dir")
        if not tmp_dir:
            tmp_dir = LocalDirectoryTarget(is_tmp=True)
        elif not isinstance(tmp_dir, LocalDirectoryTarget):
            tmp_dir = LocalDirectoryTarget(str(tmp_dir))
        kwargs["tmp_dir"] = tmp_dir

        # enter localize contexts of all targets
        with localize_file_targets(self.targets, *args, **kwargs) as localized_targets:
            # create a copy of this collection that wraps the localized targets
            yield self.__class__(localized_targets, **self._copy_kwargs())


class SiblingFileCollectionBase(FileCollection):
    """
    Base class for file collections whose elements are located in the same directory (siblings).
    """

    @classmethod
    def _exists_in_basenames(cls, target, basenames, optional_existing, target_dirs):
        if optional_existing not in (None, no_value) and target.optional:
            return optional_existing
        if isinstance(target, SiblingFileCollectionBase):
            return target._exists_fwd(
                basenames=basenames,
                optional_existing=optional_existing,
            )
        if isinstance(target, TargetCollection):
            return target.exists(exists_func=partial(
                cls._exists_in_basenames,
                basenames=basenames,
                optional_existing=optional_existing,
                target_dirs=target_dirs,
            ))
        if isinstance(basenames, dict):
            if target_dirs and target in target_dirs:
                basenames = basenames[target_dirs[target]]
            else:
                # need to find find the collection manually, that could possibly contain the target,
                # then use its basenames
                for col_absdir, _basenames in basenames.items():
                    if _target_path_in_dir(target, col_absdir):
                        basenames = _basenames
                        break
                else:
                    return False
        if not basenames:
            return False
        return target.basename in basenames

    def remove(self, silent=True):
        for targets in self.iter_existing(unpack=False):
            for t in targets:
                t.remove(silent=silent)


class SiblingFileCollection(SiblingFileCollectionBase):
    """
    Collection of targets that represent files which are all located in the same directory.
    Specifically, the performance of :py:meth:`exists` and :py:meth:`count` can greatly improve with
    respect to the standard :py:class:`FileCollection` as the directory listing is used internally.
    This is especially useful for large collections of remote files.
    """

    @classmethod
    def from_directory(cls, directory, **kwargs):
        # dir should be a FileSystemDirectoryTarget or a string, in which case it is interpreted as
        # a local path
        if isinstance(directory, FileSystemDirectoryTarget):
            d = directory
        elif directory:
            d = LocalDirectoryTarget(str(directory))
        else:
            raise TypeError("directory must either be a string or a FileSystemDirectoryTarget "
                "object, got '{}'".format(directory))

        # find all files, pass kwargs which may filter the result further
        kwargs["type"] = "f"
        basenames = d.listdir(**kwargs)

        # convert to file targets
        targets = [d.child(basename, type="f") for basename in basenames]

        return cls(targets)

    def __init__(self, *args, **kwargs):
        SiblingFileCollectionBase.__init__(self, *args, **kwargs)

        # find the first target and store its directory
        if self.first_target is None:
            raise Exception("{} requires at least one file target".format(self.__class__.__name__))
        self.dir = self.first_target.parent

        # check that targets are in fact located in the same directory
        for t in flatten_collections(self._flat_target_list):
            if not _target_path_in_dir(t, self.dir):
                raise Exception("{} is not located in common directory {}".format(t, self.dir))

    def _repr_pairs(self):
        expand = Config.instance().get_expanded_bool("target", "expand_path_repr")
        dir_path = self.dir.path if expand else self.dir.unexpanded_path
        return TargetCollection._repr_pairs(self) + [("fs", self.dir.fs.name), ("dir", dir_path)]

    def _iter_state(
        self,
        existing=True,
        optional_existing=no_value,
        basenames=None,
        keys=False,
        unpack=True,
        exists_func=None,
    ):
        # the directory must exist
        if not self.dir.exists():
            return

        existing = bool(existing)
        if optional_existing is no_value:
            optional_existing = self.optional_existing

        # get all basenames
        if basenames is None:
            basenames = self.dir.listdir() if self.dir.exists() else []
        # convert to set for faster lookup
        basenames = set(basenames) if basenames else set()

        # helper to check for existence
        if exists_func is None:
            exists_func = partial(
                self._exists_in_basenames,
                basenames=basenames,
                optional_existing=optional_existing,
                target_dirs=None,
            )

        # loop and yield
        for key, targets in self._iter_flat():
            state = all(map(exists_func, targets))
            if state is existing:
                if unpack:
                    targets = self.targets[key]
                yield (key, targets) if keys else targets

    def _exists_fwd(self, **kwargs):
        fwd = ["optional_existing", "basenames", "exists_func"]
        return self.exists(**{key: kwargs[key] for key in fwd if key in kwargs})


class NestedSiblingFileCollection(SiblingFileCollectionBase):
    """
    Collection of targets that represent files which are located across several directories, with
    files in the same directory being wrapped by a :py:class:`SiblingFileCollection` to exploit its
    benefit over the standard :py:class:`FileCollection` (see description above). This is especially
    useful for large collections of remote files that are located in different (sub) directories.

    The constructor identifies targets located in the same physical directory (identified by URI),
    creates one collection for each of them, and stores them in the *collections* attribute. Key
    access, iteration, etc., is identical to the standard :py:class:`FileCollection`.
    """

    def __init__(self, *args, **kwargs):
        super(NestedSiblingFileCollection, self).__init__(*args, **kwargs)

        # as per FileCollection's init, targets are already stored in both the _flat_targets and
        # _flat_target_list attributes, but store them again in sibling file collections to speed up
        # some methods by grouping them into targets in the same physical directory
        self.collections = []
        self._flat_target_dirs = {}
        grouped_targets = defaultdict(list)
        for t in flatten_collections(self._flat_target_list):
            grouped_targets[t.parent.uri()].append(t)
        for targets in grouped_targets.values():
            # create and store the collection
            collection = SiblingFileCollection(targets)
            self.collections.append(collection)
            # remember the absolute collection dir per target for faster loopups later
            for t in targets:
                self._flat_target_dirs[t] = collection.dir.abspath

    def _repr_pairs(self):
        return SiblingFileCollectionBase._repr_pairs(self) + [("collections", len(self.collections))]

    def _iter_state(
        self,
        existing=True,
        optional_existing=no_value,
        basenames=None,
        keys=False,
        unpack=True,
        exists_func=None,
    ):
        existing = bool(existing)
        if optional_existing is no_value:
            optional_existing = self.optional_existing

        # get all basenames
        if basenames is None:
            basenames = {
                col.dir.abspath: (col.dir.listdir() if col.dir.exists() else [])
                for col in self.collections
            }
        # convert to sets for faster lookups
        basenames = {k: (set(v) if v else set()) for k, v in basenames.items()}

        # helper to check for existence
        if exists_func is None:
            exists_func = partial(
                self._exists_in_basenames,
                basenames=basenames,
                optional_existing=optional_existing,
                target_dirs=self._flat_target_dirs,
            )

        # loop and yield
        for key, targets in self._iter_flat():
            state = all(map(exists_func, targets))
            if state is existing:
                if unpack:
                    targets = self.targets[key]
                yield (key, targets) if keys else targets

    def _exists_fwd(self, **kwargs):
        fwd = ["optional_existing", "basenames", "exists_func"]
        return self.exists(**{key: kwargs[key] for key in fwd if key in kwargs})


def _target_path_in_dir(target, directory):
    # comparisons of dirnames are transparently possible for most target classes since their
    # paths are consistent, but implement a custom check for mirrored targets
    if isinstance(target, str):
        target_absdir = target
    else:
        target_absdir = (
            target.remote_target
            if isinstance(target, MirroredTarget)
            else target
        ).absdirname
    if isinstance(directory, str):
        dir_abspath = directory
    else:
        dir_abspath = (
            directory.remote_target
            if isinstance(directory, MirroredDirectoryTarget)
            else directory
        ).abspath
    # do the comparison
    return target_absdir == dir_abspath


def flatten_collections(*targets):
    lookup = deque(flatten(targets))
    targets = []

    while lookup:
        t = lookup.popleft()
        if isinstance(t, TargetCollection):
            lookup.extendleft(t._flat_target_list)
        else:
            targets.append(t)

    return targets



# coding: utf-8

"""
Bash sandbox implementation.
"""

__all__ = ["BashSandbox"]


import os
import collections

import six

from law.config import Config
from law.sandbox.base import Sandbox
from law.util import tmp_file, interruptable_popen, quote_cmd, flatten, makedirs


class BashSandbox(Sandbox):

    sandbox_type = "bash"

    config_section_prefix = sandbox_type

    @property
    def script(self):
        return os.path.expandvars(os.path.expanduser(self.name))

    @property
    def env_cache_key(self):
        return self.script

    def get_custom_config_section_postfix(self):
        return self.name

    def create_env(self):
        # strategy: create a tempfile, let python dump its full env in a subprocess and load the
        # env file again afterwards

        # helper to write the env
        def write_env(path):
            # get the bash command
            bash_cmd = self._bash_cmd()

            # pre-setup commands
            pre_setup_cmds = self._build_pre_setup_cmds()

            # post-setup commands
            post_env = self._get_env()
            post_setup_cmds = self._build_post_setup_cmds(post_env)

            # build the python command that dumps the environment
            py_cmd = "import os,pickle;" \
                + "pickle.dump(dict(os.environ),open('{}','wb'),protocol=2)".format(path)

            # build the full command
            cmd = quote_cmd(bash_cmd + ["-c", " && ".join(flatten(
                pre_setup_cmds,
                "source \"{}\" \"\"".format(self.script),  # noqa: Q003
                post_setup_cmds,
                quote_cmd(["python", "-c", py_cmd]),
            ))])

            # run it
            returncode = interruptable_popen(cmd, shell=True, executable="/bin/bash")[0]
            if returncode != 0:
                raise Exception("{} env loading failed with exit code {}".format(
                    self, returncode))

        # helper to load the env
        def load_env(path):
            pickle_kwargs = {"encoding": "utf-8"} if six.PY3 else {}
            with open(path, "rb") as f:
                try:
                    return collections.OrderedDict(six.moves.cPickle.load(f, **pickle_kwargs))
                except Exception as e:
                    raise Exception(
                        "env deserialization of sandbox {} failed: {}".format(self, e),
                    )

        # use the cache path if set
        if self.env_cache_path:
            env_cache_path = str(self.env_cache_path)
            # write it if it does not exist yet
            if not os.path.exists(env_cache_path):
                makedirs(os.path.dirname(env_cache_path))
                write_env(env_cache_path)

            # load it
            env = load_env(env_cache_path)

        else:
            # use a temp file
            with tmp_file() as tmp:
                tmp_path = os.path.realpath(tmp[1])

                # write and load it
                write_env(tmp_path)
                env = load_env(tmp_path)

        return env

    def _bash_cmd(self):
        cmd = ["bash"]

        # login flag
        cfg = Config.instance()
        cfg_section = self.get_config_section()
        if cfg.get_expanded_bool(cfg_section, "login"):
            cmd.extend(["-l"])

        return cmd

    def cmd(self, proxy_cmd):
        # get the bash command
        bash_cmd = self._bash_cmd()

        # pre-setup commands
        pre_setup_cmds = self._build_pre_setup_cmds()

        # post-setup commands
        post_env = self._get_env()
        # add staging directories
        if self.stagein_info:
            post_env["LAW_SANDBOX_STAGEIN_DIR"] = self.stagein_info.stage_dir.path
        if self.stageout_info:
            post_env["LAW_SANDBOX_STAGEOUT_DIR"] = self.stageout_info.stage_dir.path
        post_setup_cmds = self._build_post_setup_cmds(post_env)

        # handle local scheduling within the container
        if self.force_local_scheduler():
            proxy_cmd.add_arg("--local-scheduler", "True", overwrite=True)

        # build the final command
        cmd = quote_cmd(bash_cmd + ["-c", " && ".join(flatten(
            pre_setup_cmds,
            "source \"{}\" \"\"".format(self.script),  # noqa: Q003
            post_setup_cmds,
            proxy_cmd.build(),
        ))])

        return cmd



# coding: utf-8

"""
HDF5 target formatters.
"""

__all__ = ["H5pyFormatter"]


from law.target.formatter import Formatter
from law.target.file import get_path
from law.util import no_value


class H5pyFormatter(Formatter):

    name = "h5py"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith((".hdf5", ".h5"))

    @classmethod
    def load(cls, path, *args, **kwargs):
        import h5py
        return h5py.File(get_path(path), "r", *args, **kwargs)

    @classmethod
    def dump(cls, path, *args, **kwargs):
        import h5py

        perm = kwargs.pop("perm", no_value)

        ret = h5py.File(get_path(path), "w", *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)

        return ret



# coding: utf-8

"""
Abstract defintions that enable task sandboxing.
"""

__all__ = ["Sandbox", "SandboxTask"]


import os
import sys
import shlex
from abc import ABCMeta, abstractmethod, abstractproperty
from contextlib import contextmanager
from fnmatch import fnmatch
from collections import OrderedDict

import luigi
import six

from law.config import Config
from law.task.proxy import ProxyTask, ProxyAttributeTask, ProxyCommand
from law.target.local import LocalDirectoryTarget
from law.target.collection import TargetCollection
from law.parameter import NO_STR
from law.parser import root_task
from law.util import (
    colored, is_pattern, multi_match, mask_struct, map_struct, interruptable_popen, patch_object,
    flatten,
)
from law.logger import get_logger


logger = get_logger(__name__)

_current_sandbox = os.getenv("LAW_SANDBOX", "").split(",")

_sandbox_switched = os.getenv("LAW_SANDBOX_SWITCHED", "") == "1"

_sandbox_task_id = os.getenv("LAW_SANDBOX_TASK_ID", "")

_sandbox_worker_id = os.getenv("LAW_SANDBOX_WORKER_ID", "")

_sandbox_worker_first_task_id = os.getenv("LAW_SANDBOX_WORKER_FIRST_TASK_ID", "")

_sandbox_is_root_task = os.getenv("LAW_SANDBOX_IS_ROOT_TASK", "") == "1"

_sandbox_stagein_dir = os.getenv("LAW_SANDBOX_STAGEIN_DIR", "")

_sandbox_stageout_dir = os.getenv("LAW_SANDBOX_STAGEOUT_DIR", "")


# certain values must be present in a sandbox
if _sandbox_switched:
    if not _current_sandbox or not _current_sandbox[0]:
        raise Exception("LAW_SANDBOX must not be empty in a sandbox")
    if not _sandbox_task_id:
        raise Exception("LAW_SANDBOX_TASK_ID must not be empty in a sandbox")
    if not _sandbox_worker_id:
        raise Exception("LAW_SANDBOX_WORKER_ID must not be empty in a sandbox")
    if not _sandbox_worker_first_task_id:
        raise Exception("LAW_SANDBOX_WORKER_FIRST_TASK_ID must not be empty in a sandbox")


class StageInfo(object):

    def __init__(self, targets, stage_dir, staged_targets):
        super(StageInfo, self).__init__()

        self.targets = targets
        self.stage_dir = stage_dir
        self.staged_targets = staged_targets

    def __str__(self):
        tmpl = "{}.{} object at {}:\n  targets      : {}\n  stage_dir    : {}\n  staged_targets: {}"
        return tmpl.format(
            self.__class__.__module__,
            self.__class__.__name__,
            hex(id(self)),
            self.targets,
            self.stage_dir.path,
            self.staged_targets,
        )

    def __repr__(self):
        return str(self)


class SandboxVariables(object):

    fields = ()
    eq_fields = ("name",)

    @classmethod
    def from_name(cls, name):
        if not name:
            raise ValueError("cannot create {} from empty name '{}'".format(cls.__name__, name))

        # default implementation
        return cls(name, **cls.parse_name(name))

    @classmethod
    def parse_name(cls, name):
        if not cls.fields:
            return {}

        values = {}
        for i, pair in enumerate(name.split(Sandbox.delimiter)):
            if "=" not in pair:
                if i >= len(cls.fields):
                    raise ValueError("invalid format of {} item '{}'".format(cls.__name__, pair))
                key = cls.fields[i]
                value = pair
            else:
                key, value = pair.split("=", 1)
            if key not in cls.fields:
                raise ValueError("invalid {} key '{}'".format(cls.__name__, key))
            if key in values:
                raise ValueError("duplicate {} key '{}'".format(cls.__name__, key))
            values[key] = value

        return values

    def __init__(self, name):
        super(SandboxVariables, self).__init__()

        # intended to be read-only
        self._name = str(name)

    @property
    def name(self):
        return self._name

    def __str__(self):
        return self._name

    def __eq__(self, other):
        return (
            isinstance(other, self.__class__) and
            all(getattr(self, attr) == getattr(other, attr) for attr in self.eq_fields)
        )


class Sandbox(six.with_metaclass(ABCMeta, object)):
    """
    Sandbox definition.

    The config section used by instances if this or inheriting classes constructed using
    :py:attr:`config_section_prefix` followed by ``"_sandbox"`` and optional postifxes. The minimal
    set of options in the main section are:

        - ``"stagein_dir_name"`` (usually ``"stagein"``)
        - ``"stageout_dir_name"`` (usually ``"stageout"``)
        - ``"law_executable"`` (usually ``"law"``)
    """

    delimiter = "::"

    variable_cls = SandboxVariables

    # cached envs
    _envs = {}

    @classmethod
    def check_key(cls, key, silent=False):
        # commas are not allowed since the LAW_SANDBOX env variable is allowed to contain multiple
        # comma-separated sandbox keys that need to be separated
        if "," in key:
            if silent:
                return False
            raise ValueError("invalid sandbox key format '{}'".format(key))

        return True

    @classmethod
    def split_key(cls, key):
        parts = str(key).split(cls.delimiter, 1)
        if len(parts) != 2 or any(not p.strip() for p in parts):
            raise ValueError("invalid sandbox key '{}'".format(key))

        return tuple(parts)

    @classmethod
    def join_key(cls, _type, name):
        """ join_key(type, name)
        """
        return str(_type) + cls.delimiter + str(name)

    @classmethod
    def remove_type(cls, key):
        # check for key format
        cls.check_key(key)

        # remove leading type if present
        return key.split(cls.delimiter, 1)[-1]

    @classmethod
    def create_variables(cls, name):
        return cls.variable_cls.from_name(name)

    @classmethod
    def new(cls, key, *args, **kwargs):
        # check for key format
        cls.check_key(key, silent=False)

        # split the key into the sandbox type and name
        _type, name = cls.split_key(key)

        # loop recursively through subclasses and find class that matches the sandbox_type
        classes = list(cls.__subclasses__())
        while classes:
            _cls = classes.pop(0)
            if getattr(_cls, "sandbox_type", None) == _type:
                return _cls(name, *args, **kwargs)
            classes.extend(_cls.__subclasses__())

        raise Exception("no sandbox with type '{}' found".format(_type))

    def __init__(self, name, task=None, env_cache_path=None):
        super(Sandbox, self).__init__()

        # when a task is set, it must be a SandboxTask instance
        if task and not isinstance(task, SandboxTask):
            raise TypeError("sandbox task must be a SandboxTask instance, got {}".format(task))

        self.variables = self.create_variables(name)
        self.task = task
        self.env_cache_path = (
            os.path.abspath(os.path.expandvars(os.path.expanduser(str(env_cache_path))))
            if env_cache_path
            else None
        )

        # target staging info
        self.stagein_info = None
        self.stageout_info = None

    def __str__(self):
        return self.key

    @property
    def name(self):
        return self.variables.name

    def is_active(self):
        return self.key in _current_sandbox

    @property
    def key(self):
        return self.join_key(self.sandbox_type, self.name)

    def scheduler_on_host(self):
        config = luigi.interface.core()
        return multi_match(config.scheduler_host, ["0.0.0.0", "127.0.0.1", "localhost"])

    def force_local_scheduler(self):
        return False

    @abstractproperty
    def config_section_prefix(self):
        return

    @abstractproperty
    def env_cache_key(self):
        return

    @abstractmethod
    def create_env(self):
        return

    @abstractmethod
    def cmd(self, proxy_cmd):
        return

    @property
    def env(self):
        cache_key = (self.sandbox_type, self.env_cache_key)

        if cache_key not in self._envs:
            self._envs[cache_key] = self.create_env()

        return self._envs[cache_key]

    def run(self, cmd, stdout=None, stderr=None):
        if stdout is None:
            stdout = sys.stdout
        if stderr is None:
            stderr = sys.stderr

        return interruptable_popen(
            cmd,
            shell=True,
            executable="/bin/bash",
            stdout=stdout,
            stderr=stderr,
            env=self.env,
        )

    def get_custom_config_section_postfix(self):
        return self.name

    def get_config_section(self, postfix=None):
        section = self.config_section_prefix + "_sandbox"
        if postfix:
            section += "_" + postfix

        custom_section = "{}_{}".format(section, self.get_custom_config_section_postfix())

        cfg = Config.instance()
        return custom_section if cfg.has_section(custom_section) else section

    def _get_env(self):
        # environment variables to set
        env = OrderedDict()

        # default sandboxing variables
        env["LAW_SANDBOX"] = self.key.replace("$", r"\$")
        env["LAW_SANDBOX_SWITCHED"] = "1"
        if self.task:
            env["LAW_SANDBOX_TASK_ID"] = self.task.live_task_id
            env["LAW_SANDBOX_ROOT_TASK_ID"] = root_task().task_id
            env["LAW_SANDBOX_IS_ROOT_TASK"] = str(int(self.task.is_root_task()))
            if getattr(self.task, "_worker_id", None):
                env["LAW_SANDBOX_WORKER_ID"] = self.task._worker_id
            if getattr(self.task, "_worker_first_task_id", None):
                env["LAW_SANDBOX_WORKER_FIRST_TASK_ID"] = self.task._worker_first_task_id

        # extend by variables from the config file
        cfg = Config.instance()
        section = self.get_config_section(postfix="env")
        for name, value in cfg.items(section):
            if is_pattern(name):
                names = [key for key in os.environ.keys() if fnmatch(key, name)]
            else:
                names = [name]
            for name in names:
                # when there is only a key present, i.e., no value is set,
                # get it from the current environment
                env[name] = value if value is not None else os.getenv(name, "")

        # extend by variables defined on task level
        if self.task:
            task_env = self.task.sandbox_env(env)
            if task_env:
                env.update(task_env)

        return env

    def _get_volumes(self):
        volumes = OrderedDict()

        # extend by volumes from the config file
        cfg = Config.instance()
        section = self.get_config_section(postfix="volumes")
        for hdir, cdir in cfg.items(section):
            volumes[hdir] = cdir

        # extend by volumes defined on task level
        if self.task:
            task_volumes = self.task.sandbox_volumes(volumes)
            if task_volumes:
                volumes.update(task_volumes)

        return volumes

    def _expand_volume(self, vol, bin_dir=None, python_dir=None):
        def replace(vol, name, repl):
            # warn about the deprecation of the legacy format "${name}" (until v0.1)
            var = "{{LAW_FORWARD_" + name + "}}"
            vol = vol.replace(var, repl)
            return vol

        if bin_dir:
            vol = replace(vol, "BIN", str(bin_dir))
        if python_dir:
            vol = replace(vol, "PY", str(python_dir))

        return vol

    def _build_export_commands(self, env):
        export_cmds = []

        for key, value in env.items():
            export_cmds.append("export {}=\"{}\"".format(key, value))  # noqa: Q003

        return export_cmds

    def _build_pre_setup_cmds(self, env=None):
        # commands that run before the setup is performed
        pre_setup_cmds = []

        if env:
            pre_setup_cmds.extend(self._build_export_commands(env))

        if self.task:
            pre_setup_cmds.extend(self.task.sandbox_pre_setup_cmds())

        return pre_setup_cmds

    def _build_post_setup_cmds(self, env=None):
        # commands that run before the setup is performed
        post_setup_cmds = []

        if env:
            post_setup_cmds.extend(self._build_export_commands(env))

        if self.task:
            post_setup_cmds.extend(self.task.sandbox_post_setup_cmds())

        return post_setup_cmds


class SandboxProxy(ProxyTask):

    def output(self):
        return None

    @property
    def sandbox_inst(self):
        return self.task.sandbox_inst

    def create_proxy_cmd(self):
        return ProxyCommand(
            self.task,
            exclude_task_args=self.task.exclude_params_sandbox,
            exclude_global_args=["workers"],
            executable=self.task.sandbox_law_executable(),
        )

    def run(self):
        # pre_run hook
        if callable(self.task.sandbox_pre_run):
            self.task.sandbox_pre_run()

        # create a temporary direction for file staging
        tmp_dir = LocalDirectoryTarget(is_tmp=True)
        tmp_dir.touch()

        # stage-in input files
        stagein_info = self.stagein(tmp_dir)
        if stagein_info:
            # tell the sandbox
            self.sandbox_inst.stagein_info = stagein_info
            logger.debug("configured sandbox stage-in data")

        # prepare stage-out
        stageout_info = self.prepare_stageout(tmp_dir)
        if stageout_info:
            # tell the sandbox
            self.sandbox_inst.stageout_info = stageout_info
            logger.debug("configured sandbox stage-out data")

        # create the actual command to run
        cmd = self.sandbox_inst.cmd(self.create_proxy_cmd())

        # run with log section before and after actual run call
        with self._run_context(cmd):
            code, out, err = self.sandbox_inst.run(cmd)
            if code != 0:
                raise Exception(
                    "sandbox '{}' failed with exit code {}, please see the error inside the "
                    "sandboxed context above for details".format(self.sandbox_inst.key, code),
                )

        # actual stage_out
        if stageout_info:
            self.stageout(stageout_info)

        # post_run hook
        if callable(self.task.sandbox_post_run):
            self.task.sandbox_post_run()

    def stagein(self, tmp_dir):
        # check if the stage-in dir is set
        cfg = Config.instance()
        section = self.sandbox_inst.get_config_section()
        stagein_dir_name = cfg.get_expanded(section, "stagein_dir_name")
        if not stagein_dir_name:
            return None

        # determine inputs as seen by the sandbox
        with patch_object(os, "environ", self.task.env, lock=True):
            sandbox_inputs = self.task.input()

        # get the sandbox stage-in mask
        stagein_mask = self.task.sandbox_stagein(sandbox_inputs)
        if not stagein_mask:
            return None

        # apply the mask
        sandbox_inputs = mask_struct(stagein_mask, sandbox_inputs)
        if not sandbox_inputs:
            return None

        # create the stage-in directory
        stagein_dir = tmp_dir.child(stagein_dir_name, type="d")
        stagein_dir.touch()

        # create localized sandbox input representations
        staged_inputs = create_staged_target_struct(stagein_dir, sandbox_inputs)

        # perform the actual stage-in via copying
        flat_sandbox_inputs = flatten(sandbox_inputs)
        flat_staged_inputs = flatten(staged_inputs)
        while flat_sandbox_inputs:
            sandbox_input = flat_sandbox_inputs.pop(0)
            staged_input = flat_staged_inputs.pop(0)

            if isinstance(sandbox_input, TargetCollection):
                flat_sandbox_inputs = sandbox_input._flat_target_list + flat_sandbox_inputs
                flat_staged_inputs = staged_input._flat_target_list + flat_staged_inputs
                continue

            logger.debug("stage-in {} to {}".format(sandbox_input.path, staged_input.path))
            sandbox_input.copy_to_local(staged_input)

        logger.info("staged-in {} file(s)".format(len(stagein_dir.listdir())))

        return StageInfo(sandbox_inputs, stagein_dir, staged_inputs)

    def prepare_stageout(self, tmp_dir):
        # check if the stage-out dir is set
        cfg = Config.instance()
        section = self.sandbox_inst.get_config_section()
        stageout_dir_name = cfg.get_expanded(section, "stageout_dir_name")
        if not stageout_dir_name:
            return None

        # determine outputs as seen by the sandbox
        with patch_object(os, "environ", self.task.env, lock=True):
            sandbox_outputs = self.task.output()

        # get the sandbox stage-out mask
        stageout_mask = self.task.sandbox_stageout(sandbox_outputs)
        if not stageout_mask:
            return None

        # apply the mask
        sandbox_outputs = mask_struct(stageout_mask, sandbox_outputs)
        if not sandbox_outputs:
            return None

        # create the stage-out directory
        stageout_dir = tmp_dir.child(stageout_dir_name, type="d")
        stageout_dir.touch()

        # create localized sandbox output representations
        staged_outputs = create_staged_target_struct(stageout_dir, sandbox_outputs)

        return StageInfo(sandbox_outputs, stageout_dir, staged_outputs)

    def stageout(self, stageout_info):
        # perform the actual stage-out via copying
        flat_sandbox_outputs = flatten(stageout_info.targets)
        flat_staged_outputs = flatten(stageout_info.staged_targets)
        while flat_sandbox_outputs:
            sandbox_output = flat_sandbox_outputs.pop(0)
            staged_output = flat_staged_outputs.pop(0)

            if isinstance(sandbox_output, TargetCollection):
                flat_sandbox_outputs = sandbox_output._flat_target_list + flat_sandbox_outputs
                flat_staged_outputs = staged_output._flat_target_list + flat_staged_outputs
                continue

            logger.debug("stage-out {} to {}".format(staged_output.path, sandbox_output.path))
            if staged_output.exists():
                sandbox_output.copy_from_local(staged_output)
            else:
                logger.warning(
                    "could not find output target at {} for stage-out".format(staged_output.path),
                )

        logger.info("staged-out {} file(s)".format(len(stageout_info.stage_dir.listdir())))

    @contextmanager
    def _run_context(self, cmd=None):
        def print_banner(msg, color):
            print("")
            print(colored(" {} ".format(msg).center(80, "="), color=color))
            print(colored("task   : ", color=color) + colored(self.task.task_id, style="bright"))
            print(colored("sandbox: ", color=color) + colored(self.sandbox_inst.key, style="bright"))
            print(colored(80 * "=", color=color))
            print("")

        # start banner
        print_banner("entering sandbox", "magenta")

        # log the command
        if cmd:
            logger.debug("sandbox command:\n{}".format(cmd))
        sys.stdout.flush()

        try:
            yield
        finally:
            # end banner
            print_banner("leaving sandbox", "cyan")
            sys.stdout.flush()


class SandboxTask(ProxyAttributeTask):

    sandbox = luigi.Parameter(
        default=_current_sandbox[0] or NO_STR,
        description="name of the sandbox to run the task in; default: $LAW_SANDBOX when set, "
        "otherwise empty",
    )

    allow_empty_sandbox = False
    valid_sandboxes = ["*"]

    exclude_params_sandbox = {"sandbox", "log_file"}

    def __init__(self, *args, **kwargs):
        super(SandboxTask, self).__init__(*args, **kwargs)

        # store whether sandbox objects have been setup, which is done lazily,
        # and predefine all attributes that are set by it
        self._sandbox_initialized = False
        self._effective_sandbox = None
        self._sandbox_inst = None
        self._sandbox_proxy = None

    def _initialize_sandbox(self, force=False):
        if self._sandbox_initialized and not force:
            return
        self._sandbox_initialized = True

        # reset values
        self._effective_sandbox = None
        self._sandbox_inst = None
        self._sandbox_proxy = None

        # when we are already in a sandbox, this task is placed inside it, i.e., there is no nesting
        if _sandbox_switched:
            self._effective_sandbox = _current_sandbox[0]

        # when the sandbox is set via a parameter and not hard-coded,
        # check if the value is among the valid sandboxes, otherwise determine the fallback
        elif isinstance(self.__class__.sandbox, luigi.Parameter):
            if multi_match(self.sandbox, self.valid_sandboxes, mode=any):
                self._effective_sandbox = self.sandbox
            else:
                self._effective_sandbox = self.fallback_sandbox(self.sandbox)

        # just set the effective sandbox
        else:
            self._effective_sandbox = self.sandbox

        # at this point, the sandbox must be set unless it is explicitely allowed to be empty
        if self._effective_sandbox in (None, NO_STR):
            if not self.allow_empty_sandbox:
                raise Exception("task {!r} requires the sandbox parameter to be set".format(self))
            self._effective_sandbox = NO_STR

        # create the sandbox proxy when required
        if self._effective_sandbox not in (None, NO_STR):
            sandbox_inst = Sandbox.new(self._effective_sandbox, self)
            if not sandbox_inst.is_active():
                self._sandbox_inst = sandbox_inst
                self._sandbox_proxy = SandboxProxy(task=self)
                logger.debug(
                    "created sandbox proxy instance of type '{}'".format(self._effective_sandbox),
                )

    @property
    def effective_sandbox(self):
        self._initialize_sandbox()
        return self._effective_sandbox

    @property
    def sandbox_inst(self):
        self._initialize_sandbox()
        return self._sandbox_inst

    @property
    def sandbox_proxy(self):
        self._initialize_sandbox()
        return self._sandbox_proxy

    def is_sandboxed(self):
        # returns whether the task requires no additional sandboxing, i.e., if it is already in its
        # desired sandbox
        return self.effective_sandbox == NO_STR or not self.sandbox_inst

    def is_sandboxed_task(self):
        # returns whether the task is the _one_ task whose execution is actually sandboxed
        return self.live_task_id == _sandbox_task_id

    def is_root_task(self):
        # returns whether the task is the root task of the initial "law run" invocation, potentially
        # outside a sandbox
        is_root = super(SandboxTask, self).is_root_task()
        if not _sandbox_switched:
            return is_root

        return is_root and _sandbox_is_root_task

    def _proxy_staged_input(self):
        # whether the input attribute should be forwarded to _stagein_input
        # (see get_proxy_attribute used in ProxyAttributeTask.__getattribute__)
        return _sandbox_stagein_dir and self.is_sandboxed() and self.is_sandboxed_task()

    def _proxy_staged_output(self):
        # whether the output attribute should be forwarded to _stagein_output
        # (see get_proxy_attribute used in ProxyAttributeTask.__getattribute__)
        return _sandbox_stageout_dir and self.is_sandboxed() and self.is_sandboxed_task()

    def _staged_input(self):
        from law.decorator import _is_patched_localized_method

        if not _sandbox_stagein_dir:
            raise Exception(
                "LAW_SANDBOX_STAGEIN_DIR must not be empty in a sandbox when target "
                "stage-in is required",
            )

        # get the original inputs
        input_func = self.__getattribute__("input", proxy=False)
        inputs = input_func()

        # when input_func is a patched method from a localization decorator, just return the inputs
        # since the decorator already triggered the stage-in
        if _is_patched_localized_method(input_func):
            return inputs

        # create the struct of staged inputs and apply the stage-in mask
        staged_inputs = create_staged_target_struct(_sandbox_stagein_dir, inputs)
        return mask_struct(self.sandbox_stagein(inputs), staged_inputs, replace=inputs)

    def _staged_output(self):
        from law.decorator import _is_patched_localized_method

        if not _sandbox_stageout_dir:
            raise Exception(
                "LAW_SANDBOX_STAGEOUT_DIR must not be empty in a sandbox when target "
                "stage-out is required",
            )

        # get the original outputs
        output_func = self.__getattribute__("output", proxy=False)
        outputs = output_func()

        # when output_func is a patched method from a localization decorator, just return the
        # outputs since the decorator already triggered the stage-out
        if _is_patched_localized_method(output_func):
            return outputs

        # create the struct of staged outputs and apply the stage-out mask
        staged_outputs = create_staged_target_struct(_sandbox_stageout_dir, outputs)
        return mask_struct(self.sandbox_stageout(outputs), staged_outputs, replace=outputs)

    @property
    def env(self):
        return os.environ if self.is_sandboxed() else self.sandbox_inst.env

    def fallback_sandbox(self, sandbox):
        return None

    def sandbox_user(self):
        uid, gid = os.getuid(), os.getgid()

        # check if there is a config section that defines the user and group ids
        if self.sandbox_inst:
            cfg = Config.instance()
            section = self.sandbox_inst.get_config_section()
            uid = cfg.get_expanded_int(section, "uid", default=uid)
            gid = cfg.get_expanded_int(section, "gid", default=gid)

        return uid, gid

    def sandbox_stagein(self, inputs):
        # disable stage-in by default
        return False

    def sandbox_stageout(self, outputs):
        # disable stage-out by default
        return False

    def sandbox_env(self, env):
        # additional environment variables
        return {}

    def sandbox_volumes(self, volumes):
        # additional volumes to mount
        return {}

    def sandbox_pre_setup_cmds(self):
        # list of commands that are run before the sandbox is set up
        return []

    def sandbox_post_setup_cmds(self):
        # list of commands that are run after the sandbox is set up
        return []

    def sandbox_law_executable(self):
        # law executable that is used inside the sandbox
        executable = "law"

        if self.sandbox_inst:
            section = self.sandbox_inst.get_config_section()
            executable = Config.instance().get_expanded(section, "law_executable")

        return shlex.split(executable) if executable else []

    def sandbox_pre_run(self):
        # method that is invoked before the run method of the sandbox proxy is called
        return

    def sandbox_post_run(self):
        # method that is invoked after the run method of the sandbox proxy is called
        return


def create_staged_target_struct(stage_dir, struct):
    def map_target(target):
        return create_staged_target(stage_dir, target)

    def map_collection(func, collection, **kwargs):
        staged_targets = map_struct(func, collection.targets, **kwargs)
        return collection.__class__(staged_targets, **collection._copy_kwargs())

    return map_struct(map_target, struct, custom_mappings={TargetCollection: map_collection})


def create_staged_target(stage_dir, target):
    if not isinstance(stage_dir, LocalDirectoryTarget):
        stage_dir = LocalDirectoryTarget(str(stage_dir))

    return stage_dir.child(target.unique_basename, type=target.type, **target._copy_kwargs())



# coding: utf-8

"""
Formatter classes for file targets.
"""

__all__ = ["AUTO_FORMATTER", "Formatter", "get_formatter", "find_formatters", "find_formatter"]


import os
import json
import zipfile
import gzip
import tarfile
from collections import OrderedDict

import six

from law.util import no_value, make_list, import_file
from law.logger import get_logger


logger = get_logger(__name__)


AUTO_FORMATTER = "auto"


class FormatterRegister(type):

    formatters = OrderedDict()

    def __new__(metacls, classname, bases, classdict):
        cls = type.__new__(metacls, classname, bases, classdict)

        if cls.name in metacls.formatters:
            raise ValueError("duplicate formatter name '{}' for class {}".format(cls.name, cls))
        if cls.name == AUTO_FORMATTER:
            raise ValueError("formatter class {} must not be named '{}'".format(
                cls, AUTO_FORMATTER))

        # store classes by name
        if cls.name != "_base":
            metacls.formatters[cls.name] = cls
            logger.debug("registered target formatter '{}'".format(cls.name))

        return cls


def get_formatter(name, silent=False):
    """
    Returns the formatter class whose name attribute is *name*. When no class could be found and
    *silent* is *True*, *None* is returned. Otherwise, an exception is raised.
    """
    formatter = FormatterRegister.formatters.get(name)
    if formatter or silent:
        return formatter
    raise Exception("cannot find formatter '{}'".format(name))


def find_formatters(path, mode, silent=True):
    """
    Returns a list of formatter classes which would accept the file given by *path* and *mode*,
    which should either be ``"load"`` or ``"dump"``. When no classes could be found and *silent* is
    *True*, an empty list is returned. Otherwise, an exception is raised.
    """
    path = get_path(path)
    formatters = [f for f in six.itervalues(FormatterRegister.formatters) if f.accepts(path, mode)]
    if formatters or silent:
        return formatters
    raise Exception("cannot find any '{}' formatter for {}".format(mode, path))


def find_formatter(path, mode, name=AUTO_FORMATTER):
    """
    Returns the formatter class whose name attribute is *name* when *name* is not *AUTO_FORMATTER*.
    Otherwise, the first formatter that accepts *path* is returned. Internally, this method simply
    uses :py:func:`get_formatter` or :py:func:`find_formatters` depending on the value of *name*.
    """
    if name == AUTO_FORMATTER:
        return find_formatters(path, mode, silent=False)[0]
    return get_formatter(name, silent=False)


class Formatter(six.with_metaclass(FormatterRegister, object)):

    name = "_base"

    # modes
    LOAD = "load"
    DUMP = "dump"

    @classmethod
    def accepts(cls, path, mode):
        raise NotImplementedError

    @classmethod
    def load(cls, path, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def dump(cls, path, *args, **kwargs):
        raise NotImplementedError

    @classmethod
    def chmod(cls, target, perm=None):
        if not isinstance(target, FileSystemTarget):
            return

        if perm is None:
            perm = (
                target.fs.default_file_perm
                if isinstance(target, FileSystemFileTarget)
                else target.fs.default_dir_perm
            )

        if perm:
            target.chmod(perm)


class TextFormatter(Formatter):

    name = "text"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith(".txt")

    @classmethod
    def load(cls, path, *args, **kwargs):
        with open(get_path(path), "r") as f:
            return f.read(*args, **kwargs)

    @classmethod
    def dump(cls, path, content, *args, **kwargs):
        perm = kwargs.pop("perm", no_value)

        with open(get_path(path), "w") as f:
            f.write(str(content), *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)


class JSONFormatter(Formatter):

    name = "json"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith(".json")

    @classmethod
    def load(_cls, path, *args, **kwargs):
        # kwargs might contain *cls*
        with open(get_path(path), "r") as f:
            return json.load(f, *args, **kwargs)

    @classmethod
    def dump(_cls, path, obj, *args, **kwargs):
        perm = kwargs.pop("perm", no_value)

        # kwargs might contain *cls*
        with open(get_path(path), "w") as f:
            ret = json.dump(obj, f, *args, **kwargs)

        if perm != no_value:
            _cls.chmod(path, perm)

        return ret


class PickleFormatter(Formatter):

    name = "pickle"

    @classmethod
    def accepts(cls, path, mode):
        path = get_path(path)
        return path.endswith((".pkl", ".pickle", ".p"))

    @classmethod
    def load(cls, path, *args, **kwargs):
        with open(get_path(path), "rb") as f:
            return six.moves.cPickle.load(f, *args, **kwargs)

    @classmethod
    def dump(cls, path, obj, *args, **kwargs):
        perm = kwargs.pop("perm", no_value)

        with open(get_path(path), "wb") as f:
            ret = six.moves.cPickle.dump(obj, f, *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)

        return ret


class YAMLFormatter(Formatter):

    name = "yaml"

    @classmethod
    def accepts(cls, path, mode):
        path = get_path(path)
        return path.endswith((".yaml", ".yml"))

    @classmethod
    def load(cls, path, *args, **kwargs):
        import yaml

        with open(get_path(path), "r") as f:
            return yaml.safe_load(f, *args, **kwargs)

    @classmethod
    def dump(cls, path, obj, *args, **kwargs):
        import yaml

        perm = kwargs.pop("perm", no_value)

        with open(get_path(path), "w") as f:
            ret = yaml.dump(obj, f, *args, **kwargs)

        if perm != no_value:
            cls.chmod(path, perm)

        return ret


class ZipFormatter(Formatter):

    name = "zip"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith(".zip")

    @classmethod
    def load(cls, path, dst, *args, **kwargs):
        # assume read mode, but also check args and kwargs
        mode = "r"
        if args:
            mode = args[0]
            args = args[1:]
        elif "mode" in kwargs:
            mode = kwargs.pop("mode")

        # arguments passed to extractall()
        extractall_kwargs = kwargs.pop("extractall_kwargs", None) or {}

        # open zip file and extract to dst
        with zipfile.ZipFile(get_path(path), mode, *args, **kwargs) as f:
            f.extractall(get_path(dst), **extractall_kwargs)

    @classmethod
    def dump(cls, path, src, *args, **kwargs):
        # assume write mode, but also check args and kwargs
        mode = "w"
        if args:
            mode = args[0]
            args = args[1:]
        elif "mode" in kwargs:
            mode = kwargs.pop("mode")

        perm = kwargs.pop("perm", no_value)

        # arguments passed to write()
        write_kwargs = kwargs.pop("write_kwargs", None) or {}

        # open a new zip file and add all files in src
        with zipfile.ZipFile(get_path(path), mode, *args, **kwargs) as f:
            src = get_path(src)
            if os.path.isfile(src):
                f.write(src, os.path.basename(src), **write_kwargs)
            else:
                for elem in os.listdir(src):
                    f.write(os.path.join(src, elem), elem, **write_kwargs)

        if perm != no_value:
            cls.chmod(path, perm)


class GZipFormatter(Formatter):

    name = "gzip"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith(".gz")

    @classmethod
    def load(cls, path, *args, **kwargs):
        # assume read mode, but also check args and kwargs
        mode = "r"
        if args:
            mode = args[0]
            args = args[1:]
        elif "mode" in kwargs:
            mode = kwargs.pop("mode")

        # arguments passed to read()
        read_kwargs = kwargs.pop("read_kwargs", None) or {}

        # open with gzip and return content
        with gzip.open(get_path(path), mode, *args, **kwargs) as f:
            return f.read(**read_kwargs)

    @classmethod
    def dump(cls, path, obj, *args, **kwargs):
        # assume write mode, but also check args and kwargs
        mode = "w"
        if args:
            mode = args[0]
            args = args[1:]
        elif "mode" in kwargs:
            mode = kwargs.pop("mode")

        perm = kwargs.pop("perm", no_value)

        # arguments passed to write()
        write_kwargs = kwargs.pop("write_kwargs", None) or {}

        # write into a new gzip file
        with gzip.open(get_path(path), mode, *args, **kwargs) as f:
            ret = f.write(obj, **write_kwargs)

        if perm != no_value:
            cls.chmod(path, perm)

        return ret


class TarFormatter(Formatter):

    name = "tar"

    @classmethod
    def infer_compression(cls, path):
        path = get_path(path)
        if path.endswith((".tar.gz", ".tgz")):
            return "gz"
        if path.endswith((".tar.bz2", ".tbz2", ".bz2")):
            return "bz2"
        if path.endswith((".tar.xz", ".txz", ".lzma")):
            return "xz"
        return None

    @classmethod
    def accepts(cls, path, mode):
        return cls.infer_compression(path) is not None

    @classmethod
    def load(cls, path, dst, *args, **kwargs):
        # get the mode from args and kwargs, default to read mode with inferred compression
        if args:
            mode = args[0]
            args = args[1:]
        elif "mode" in kwargs:
            mode = kwargs.pop("mode")
        else:
            compression = cls.infer_compression(path)
            mode = "r" if not compression else "r:" + compression

        # arguments passed to extractall()
        extractall_kwargs = kwargs.pop("extractall_kwargs", None) or {}

        # open zip file and extract to dst
        with tarfile.open(get_path(path), mode, *args, **kwargs) as f:
            f.extractall(get_path(dst), **extractall_kwargs)

    @classmethod
    def dump(cls, path, src, *args, **kwargs):
        # get the mode from args and kwargs, default to write mode with inferred compression
        if args:
            mode = args[0]
            args = args[1:]
        elif "mode" in kwargs:
            mode = kwargs.pop("mode")
        else:
            compression = cls.infer_compression(path)
            mode = "w" if not compression else "w:" + compression

        perm = kwargs.pop("perm", no_value)

        # arguments passed to add()
        add_kwargs = kwargs.pop("add_kwargs", None) or {}

        # backwards compatibility
        _filter = kwargs.pop("filter", None)
        if _filter is not None:
            logger.warning_once(
                "passing filter=callback' to TarFormatter.dump is deprecated and will be removed "
                "in a future release; please use 'add_kwargs=dict(filter=callback)' instead",
            )
            add_kwargs["filter"] = _filter

        # open a new zip file and add all files in src
        with tarfile.open(get_path(path), mode, *args, **kwargs) as f:
            srcs = [os.path.abspath(get_path(src)) for src in make_list(src)]
            common_prefix = os.path.commonprefix(srcs)
            for src in srcs:
                _add_kwargs = {"arcname": os.path.relpath(src, common_prefix)}
                _add_kwargs.update(add_kwargs)
                f.add(src, **_add_kwargs)

        if perm != no_value:
            cls.chmod(path, perm)


class PythonFormatter(Formatter):

    name = "python"

    @classmethod
    def accepts(cls, path, mode):
        return get_path(path).endswith(".py")

    @classmethod
    def load(cls, path, *args, **kwargs):
        return import_file(get_path(path), *args, **kwargs)


# trailing imports
from law.target.file import get_path, FileSystemTarget, FileSystemFileTarget



# coding: utf-8

"""
Target classes that represent remote files and directories and have a local, optionally read-only
mirror (e.g. through a local mount of the remote file system).
"""

__all__ = ["MirroredTarget", "MirroredFileTarget", "MirroredDirectoryTarget"]


import os
import contextlib
import time
import threading

from law.config import Config
from law.target.file import FileSystemTarget, FileSystemFileTarget, FileSystemDirectoryTarget
from law.target.local import LocalFileTarget, LocalDirectoryTarget
from law.target.remote import RemoteFileTarget, RemoteDirectoryTarget
from law.util import patch_object


local_root_check_lock = threading.Lock()


class MirroredTarget(FileSystemTarget):

    _existing_local_roots = {}

    @classmethod
    def check_local_root(cls, path, depth=1):
        path = str(path)

        # path must start with a separator
        if path == os.sep:
            return True
        if not path or not path.startswith(os.sep):
            return False

        # get the root path or mount point of the path (e.g. "/mnt")
        root_path = os.sep.join(path.split(os.sep)[:depth + 1])
        if root_path not in cls._existing_local_roots:
            with local_root_check_lock:
                cls._existing_local_roots[root_path] = os.path.exists(root_path)

        return cls._existing_local_roots[root_path]

    def __init__(
        self,
        path,
        _is_file,
        remote_target=None,
        remote_target_cls=None,
        remote_fs=None,
        remote_kwargs=None,
        local_target=None,
        local_fs=None,
        local_kwargs=None,
        local_read_only=True,
        local_sync=True,
        **kwargs,
    ):
        path = path.lstrip(os.sep)

        # create a remote target
        if not remote_target:
            if not remote_fs:
                raise ValueError("either remote_target or remote_fs must be given")
            if not remote_target_cls:
                raise ValueError("either remote_target or remote_target_cls must be given")
            if _is_file and not issubclass(remote_target_cls, RemoteFileTarget):
                raise TypeError(
                    "remote_target_cls must be a subclass of RemoteFileTarget: {}".format(
                        remote_target_cls,
                    ),
                )
            if not _is_file and not issubclass(remote_target_cls, RemoteDirectoryTarget):
                raise TypeError(
                    "remote_target_cls must be a subclass of RemoteDirectoryTarget: {}".format(
                        remote_target_cls,
                    ),
                )
            remote_kwargs = remote_kwargs.copy() if remote_kwargs else {}
            remote_kwargs["fs"] = remote_fs
            remote_target = remote_target_cls(os.sep + path, **remote_kwargs)
        else:
            if _is_file and not isinstance(remote_target, RemoteFileTarget):
                raise TypeError(
                    "remote_target must be an instance of RemoteFileTarget: {}".format(
                        remote_target,
                    ),
                )
            if not _is_file and not isinstance(remote_target, RemoteDirectoryTarget):
                raise TypeError(
                    "remote_target must be an instance of RemoteDirectoryTarget: {}".format(
                        remote_target,
                    ),
                )

        # create a local target
        if not local_target:
            local_kwargs = local_kwargs.copy() if local_kwargs else {}
            if not local_fs:
                local_fs = Config.instance().get_expanded("target", "default_local_fs")
            local_kwargs["fs"] = local_fs
            local_target_cls = LocalFileTarget if _is_file else LocalDirectoryTarget
            local_target = local_target_cls(path, **local_kwargs)
        else:
            if _is_file and not isinstance(local_target, LocalFileTarget):
                raise TypeError(
                    "local_target must be an instance of LocalFileTarget: {}".format(
                        local_target,
                    ),
                )
            if not _is_file and not isinstance(local_target, LocalDirectoryTarget):
                raise TypeError(
                    "local_target must be an instance of LocalDirectoryTarget: {}".format(
                        local_target,
                    ),
                )

        # store targets
        self.remote_target = remote_target
        self.local_target = local_target

        # additional attributes
        self.local_read_only = local_read_only
        self.local_sync = local_sync

        # temporary, forced file system
        self._force_fs = None

        super().__init__(path, **kwargs)

    @property
    def _local_root_depth(self):
        return self.local_target.fs.local_root_depth

    def _local_root_exists(self):
        return self.check_local_root(self.local_target.abspath, depth=self._local_root_depth)

    def _local_target_exists(self, *args, **kwargs):
        return self._local_root_exists() and self.local_target.exists(*args, **kwargs)

    def _parent_args(self):
        parent_kwargs = {
            "remote_target": self.remote_target.parent,
            "local_target": self.local_target.parent,
            "local_read_only": self.local_read_only,
            "local_sync": self.local_sync,
        }
        return (), parent_kwargs

    def _wait_for_local(self, missing=False, timeout=0.5, attempts=90):
        if not self.check_local_root(self.local_target.abspath, depth=self._local_root_depth):
            return

        sleep_counter = 0
        while self.local_target.exists() == missing:
            time.sleep(timeout)
            sleep_counter += 1

            if sleep_counter >= attempts:
                state = "disappear" if missing else "exist"
                raise Exception(
                    "timeout while waiting for local representation {!r} to {}".format(
                        self.local_target, state,
                    ),
                )

    @property
    def fs(self):
        if self._force_fs is not None:
            return self._force_fs

        return (
            self.local_target.fs
            if self._local_target_exists()
            else self.remote_target.fs
        )

    @contextlib.contextmanager
    def force_fs(self, fs):
        with patch_object(self, "_force_fs", fs):
            yield

    @property
    def abspath(self):
        return (
            self.local_target.abspath
            if self._local_target_exists()
            else self.remote_target.abspath
        )

    @property
    def dirname(self):
        return (
            self.local_target.dirname
            if self._local_target_exists()
            else self.remote_target.dirname
        )

    @property
    def absdirname(self):
        return (
            self.local_target.absdirname
            if self._local_target_exists()
            else self.remote_target.absdirname
        )

    @property
    def basename(self):
        return self.local_target.basename

    def stat(self, *args, **kwargs):
        return (
            self.local_target.stat(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.stat(*args, **kwargs)
        )

    def exists(self, *args, **kwargs):
        return (
            self.local_target.exists(*args, **kwargs) or
            self.remote_target.exists(*args, **kwargs)
        )

    def remove(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.remove(*args, **kwargs)

        ret = self.remote_target.remove(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=True)
        return ret

    def chmod(self, *args, **kwargs):
        return self.remote_target.chmod(*args, **kwargs)

    def uri(self, *args, **kwargs):
        return (
            self.local_target.uri(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.uri(*args, **kwargs)
        )

    def copy_to(self, *args, **kwargs):
        return (
            self.local_target.copy_to(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.copy_to(*args, **kwargs)
        )

    def copy_from(self, *args, **kwargs):
        ret = self.remote_target.copy_from(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=False)
        return ret

    def move_to(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.move_to(*args, **kwargs)

        ret = self.remote_target.move_to(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=True)
        return ret

    def move_from(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.move_from(*args, **kwargs)

        ret = self.remote_target.move_from(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=False)
        return ret

    def copy_to_local(self, *args, **kwargs):
        return (
            self.local_target.copy_to_local(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.copy_to_local(*args, **kwargs)
        )

    def copy_from_local(self, *args, **kwargs):
        ret = self.remote_target.copy_from_local(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=False)
        return ret

    def move_to_local(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.move_to_local(*args, **kwargs)

        ret = self.remote_target.move_to_local(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=True)
        return ret

    def move_from_local(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.move_from_local(*args, **kwargs)

        ret = self.remote_target.move_from_local(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=False)
        return ret

    @contextlib.contextmanager
    def localize(self, mode="r", **kwargs):
        with (
            self.local_target.localize(mode, **kwargs)
            if (mode == "r" or not self.local_read_only) and self._local_target_exists()
            else self.remote_target.localize(mode, **kwargs)
        ) as ret:
            yield ret

        if mode == "w" and self.local_read_only and self.local_sync:
            self._wait_for_local(missing=False)

    def touch(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.touch(*args, **kwargs)

        ret = self.remote_target.touch(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=False)
        return ret

    def load(self, *args, **kwargs):
        return (
            self.local_target.load(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.load(*args, **kwargs)
        )

    def dump(self, *args, **kwargs):
        if not self.local_read_only:
            return self.local_target.dump(*args, **kwargs)

        ret = self.remote_target.dump(*args, **kwargs)
        if self.local_sync:
            self._wait_for_local(missing=False)
        return ret


class MirroredFileTarget(FileSystemFileTarget, MirroredTarget):

    def __init__(self, path, **kwargs):
        super().__init__(path, _is_file=True, **kwargs)

    @contextlib.contextmanager
    def open(self, mode, **kwargs):
        with (
            self.local_target.open(mode, **kwargs)
            if (mode == "r" or not self.local_read_only) and self._local_target_exists()
            else self.remote_target.open(mode, **kwargs)
        ) as ret:
            yield ret

        if mode == "w" and self.local_read_only and self.local_sync:
            self._wait_for_local(missing=False)


class MirroredDirectoryTarget(FileSystemDirectoryTarget, MirroredTarget):

    def __init__(self, path, **kwargs):
        super().__init__(path, _is_file=False, **kwargs)

    def _child_args(self, path, type):
        child_kwargs = {
            "remote_target": self.remote_target.child(path, type=type),
            "local_target": self.local_target.child(path, type=type),
            "local_read_only": self.local_read_only,
            "local_sync": self.local_sync,
        }
        return (), child_kwargs

    def listdir(self, *args, **kwargs):
        return (
            self.local_target.listdir(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.listdir(*args, **kwargs)
        )

    def glob(self, *args, **kwargs):
        return (
            self.local_target.glob(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.glob(*args, **kwargs)
        )

    def walk(self, *args, **kwargs):
        return (
            self.local_target.walk(*args, **kwargs)
            if self._local_target_exists()
            else self.remote_target.walk(*args, **kwargs)
        )


MirroredTarget.file_class = MirroredFileTarget
MirroredTarget.directory_class = MirroredDirectoryTarget



# coding: utf-8

"""
Mercurial-related tasks.
"""

__all__ = ["BundleMercurialRepository"]


import os
import threading
import subprocess
from abc import abstractmethod

import luigi

from law.task.base import Task
from law.target.file import get_path
from law.target.local import LocalFileTarget
from law.parameter import NO_STR, CSVParameter
from law.decorator import log
from law.util import rel_path, interruptable_popen, quote_cmd


class BundleMercurialRepository(Task):

    task_namespace = "law.mercurial"

    exclude_files = CSVParameter(
        default=(),
        description="patterns of files to exclude, default: ()",
    )
    include_files = CSVParameter(
        default=(),
        description="patterns of files to force-include, takes precedence over .hgignore, "
        "default: ()",
    )
    custom_checksum = luigi.Parameter(
        default=NO_STR,
        description="a custom checksum to use, default: NO_STR",
    )

    def __init__(self, *args, **kwargs):
        super(BundleMercurialRepository, self).__init__(*args, **kwargs)

        self._checksum = None
        self._checksum_lock = threading.RLock(())

    @abstractmethod
    def get_repo_path(self):
        return

    @property
    def checksum(self):
        if self.custom_checksum != NO_STR:
            return self.custom_checksum

        with self._checksum_lock:
            if self._checksum is None:
                cmd = quote_cmd([
                    rel_path(__file__, "scripts", "repository_checksum.sh"),
                    get_path(self.get_repo_path()),
                ])

                code, out, _ = interruptable_popen(
                    cmd,
                    shell=True,
                    executable="/bin/bash",
                    stdout=subprocess.PIPE,
                )
                if code != 0:
                    raise Exception("repository checksum calculation failed")

                self._checksum = out.strip()

        return self._checksum

    def output(self):
        repo_base = os.path.basename(get_path(self.get_repo_path()))
        repo_base = os.path.abspath(os.path.expandvars(os.path.expanduser(repo_base)))
        return LocalFileTarget("{}_{}.tgz".format(repo_base, self.checksum))

    @log
    def run(self):
        with self.output().localize("w") as tmp:
            self.bundle(tmp.path)

    def bundle(self, dst_path):
        bundle_script = rel_path(__file__, "scripts", "bundle_repository.sh")
        cmd = [bundle_script, get_path(self.get_repo_path()), get_path(dst_path)]
        cmd += [" ".join(self.exclude_files)]
        cmd += [" ".join(self.include_files)]

        code = interruptable_popen(cmd, executable="/bin/bash")[0]
        if code != 0:
            raise Exception("repository bundling failed")



# coding: utf-8

"""
Base classes for implementing remote job management and job file creation.
"""

__all__ = ["BaseJobManager", "BaseJobFileFactory", "JobArguments", "JobInputFile"]


import os
import time
import shutil
import tempfile
import fnmatch
import base64
import copy
import re
import json
from collections import defaultdict
from multiprocessing.pool import ThreadPool
from threading import Lock
from abc import ABCMeta, abstractmethod

import six

from law.config import Config
from law.target.file import get_scheme, get_path
from law.target.remote.base import RemoteTarget
from law.util import (
    colored, make_list, make_tuple, iter_chunks, makedirs, create_hash, create_random_string,
    increment_path,
)
from law.logger import get_logger


logger = get_logger(__name__)


def get_async_result_silent(result, timeout=None):
    """
    Calls the ``get([timeout])`` method of an `AsyncResult
    <https://docs.python.org/latest/library/multiprocessing.html#multiprocessing.pool.AsyncResult>`__
    object *result* and returns its value. The only difference is that potentially raised exceptions
    are returned instead of re-raised.

    """  # noqa
    try:
        return result.get(timeout)
    except Exception as e:
        return e


class BaseJobManager(six.with_metaclass(ABCMeta, object)):
    """
    Base class that defines how remote jobs are submitted, queried, cancelled and cleaned up. It
    also defines the most common job states:

    - PENDING: The job is submitted and waiting to be processed.
    - RUNNUNG: The job is running.
    - FINISHED: The job is completed and successfully finished.
    - RETRY: The job is completed but failed. It can be resubmitted.
    - FAILED: The job is completed but failed. It cannot or should not be recovered.

    The particular job manager implementation should match its own, native states to these common
    states.

    *status_names* and *status_diff_styles* are used in :py:meth:`status_line` and default to
    :py:attr:`default_status_names` and :py:attr:`default_status_diff_styles`. *threads* is the
    default number of concurrent threads that are used in :py:meth:`submit_batch`,
    :py:meth:`cancel_batch`, :py:meth:`cleanup_batch` and :py:meth:`query_batch`.

    .. py:classattribute:: PENDING

        type: string

        Flag that represents the ``PENDING`` status.

    .. py:classattribute:: RUNNING

        type: string

        Flag that represents the ``RUNNING`` status.

    .. py:classattribute:: FINISHED

        type: string

        Flag that represents the ``FINISHED`` status.

    .. py:classattribute:: RETRY

        type: string

        Flag that represents the ``RETRY`` status.

    .. py:classattribute:: FAILED

        type: string

        Flag that represents the ``FAILED`` status.

    .. py:classattribute:: default_status_names

        type: list

        The list of all default status flags that is used in :py:meth:`status_line`.

    .. py:classattribute:: default_status_diff_styles

        type: dict

        A dictionary that defines to coloring styles per job status that is used in
        :py:meth:`status_line`.

    .. py:classattribute:: job_grouping_submit

        type: bool

        Whether this manager implementation groups jobs into single interactions for submission. In
        general, this means that the submission of a single job file can result in multiple jobs on
        the remote batch system.

    .. py:classattribute:: job_grouping_cancel

        type: bool

        Whether this manager implementation groups jobs into single interactions for cancelling
        jobs.

    .. py:classattribute:: job_grouping_cleanup

        type: bool

        Whether this manager implementation groups jobs into single interactions for cleaning up
        jobs.

    .. py:classattribute:: job_grouping_query

        type: bool

        Whether this manager implementation groups jobs into single interactions for querying job
        statuses.

    .. py:classattribute:: chunk_size_submit

        type: int

        The default chunk size value when no value is given in :py:meth:`submit_batch`. If the value
        evaluates to *False*, no chunking is allowed.


    .. py:classattribute:: chunk_size_cancel

        type: int

        The default chunk size value when no value is given in :py:meth:`cancel_batch`. If the value
        evaluates to *False*, no chunking is allowed.

    .. py:classattribute:: chunk_size_cleanup

        type: int

        The default chunk size value when no value is given in :py:meth:`cleanup_batch`. If the
        value evaluates to *False*, no chunking is allowed.

    .. py:classattribute:: chunk_size_query

        type: int

        The default chunk size value when no value is given in :py:meth:`query_batch`. If the value
        evaluates to *False*, no chunking is allowed.
    """

    PENDING = "pending"
    RUNNING = "running"
    FINISHED = "finished"
    RETRY = "retry"
    FAILED = "failed"

    default_status_names = [PENDING, RUNNING, FINISHED, RETRY, FAILED]

    # color styles per status when job count decreases / stagnates / increases
    default_status_diff_styles = {
        PENDING: ({}, {}, {"color": "green"}),
        RUNNING: ({}, {}, {"color": "green"}),
        FINISHED: ({}, {}, {"color": "green"}),
        RETRY: ({"color": "green"}, {}, {"color": "red"}),
        FAILED: ({}, {}, {"color": "red", "style": "bright"}),
    }

    # job grouping settings per method
    job_grouping_submit = False
    job_grouping_cancel = False
    job_grouping_cleanup = False
    job_grouping_query = False

    # chunking settings for unbatched methods
    # disabled by default
    chunk_size_submit = 0
    chunk_size_cancel = 0
    chunk_size_cleanup = 0
    chunk_size_query = 0

    @classmethod
    def job_status_dict(cls, job_id=None, status=None, code=None, error=None, extra=None):
        """
        Returns a dictionay that describes the status of a job given its *job_id*, *status*, return
        *code*, *error*, and additional *extra* data.
        """
        return dict(job_id=job_id, status=status, code=code, error=error, extra=extra)

    @classmethod
    def cast_job_id(cls, job_id):
        """
        Hook for casting an input *job_id*, for instance, after loading serialized data from json.
        """
        return job_id

    def __init__(self, status_names=None, status_diff_styles=None, threads=1):
        super(BaseJobManager, self).__init__()

        self.status_names = status_names or list(self.default_status_names)
        self.status_diff_styles = status_diff_styles or self.default_status_diff_styles.copy()
        self.threads = threads

        self.last_counts = [0] * len(self.status_names)

    @abstractmethod
    def submit(self):
        """
        Abstract atomic or group job submission.
        Can throw exceptions.
        Should return a list of job ids.
        """
        return

    @abstractmethod
    def cancel(self):
        """
        Abstract atomic or group job cancellation.
        Can throw exceptions.
        Should return a dictionary mapping job ids to per-job return values.
        """
        return

    @abstractmethod
    def cleanup(self):
        """
        Abstract atomic or group job cleanup.
        Can throw exceptions.
        Should return a dictionary mapping job ids to per-job return values.
        """
        return

    @abstractmethod
    def query(self):
        """
        Abstract atomic or group job status query.
        Can throw exceptions.
        Should return a dictionary mapping job ids to per-job return values.
        """
        return

    def group_job_ids(self, job_ids):
        """
        Hook that needs to be implemented if the job mananger supports grouping of jobs, i.e., when
        :py:attr:`job_grouping_submit`, :py:attr:`job_grouping_query`, etc. is *True*, and
        potentially used during status queries, job cancellation and removal. If so, it should take
        a sequence of *job_ids* and return a dictionary mapping ids of group jobs (used for queries
        etc) to the corresponding lists of original job ids, with an arbitrary grouping mechanism.
        """
        raise NotImplementedError(
            "internal error, {}.group_job_ids not implemented".format(self.__class__.__name__),
        )

    def _apply_batch(
        self,
        func,
        result_type,
        job_objs,
        default_chunk_size,
        threads=None,
        chunk_size=None,
        callback=None,
        **kwargs  # noqa
    ):
        # default arguments
        threads = max(threads or self.threads or 1, 1)

        # is chunking allowed?
        chunk_size = max(chunk_size or default_chunk_size, 0) if default_chunk_size else 0
        chunking = chunk_size > 0

        # build chunks if needed
        job_objs = make_list(job_objs)
        job_objs = list(iter_chunks(job_objs, chunk_size)) if chunking else job_objs

        # factory to call the passed callback for each job file even when chunking
        def cb_factory(i):
            if not callable(callback):
                return None

            if chunking:
                def wrapper(result_data):
                    offset = sum(map(len, job_objs[:i]))
                    for j in range(len(job_objs[i])):
                        data = result_data if isinstance(result_data, Exception) else result_data[j]
                        callback(offset + j, data)
            else:
                def wrapper(data):
                    callback(i, data)

            return wrapper

        # threaded processing
        pool = ThreadPool(threads)
        results = [
            pool.apply_async(func, (arg,), kwargs, callback=cb_factory(i))
            for i, arg in enumerate(job_objs)
        ]
        pool.close()
        pool.join()

        # store result data or an exception
        result_data = result_type()
        if chunking:
            for _job_objs, res in six.moves.zip(job_objs, results):
                data = get_async_result_silent(res)
                for i, job_obj in enumerate(_job_objs):
                    if isinstance(result_data, list):
                        result_data.append(data if isinstance(data, Exception) else data[i])
                    else:
                        result_data[job_obj] = data if isinstance(data, Exception) else data[job_obj]
        else:
            for job_obj, res in six.moves.zip(job_objs, results):
                data = get_async_result_silent(res)
                if isinstance(result_data, list):
                    result_data.append(data)
                else:
                    result_data[job_obj] = data

        return result_data

    def submit_batch(self, job_files, threads=None, chunk_size=None, callback=None, **kwargs):
        """
        Submits a batch of jobs given by *job_files* via a thread pool of size *threads* which
        defaults to its instance attribute. When *chunk_size*, which defaults to
        :py:attr:`chunk_size_submit`, is not negative, *job_files* are split into chunks of that
        size which are passed to :py:meth:`submit`.

        When *callback* is set, it is invoked after each successful job submission with the index of
        the corresponding job file (starting at 0) and either the assigned job id or an exception if
        any occurred. All other *kwargs* are passed to
        :py:meth:`submit`.

        The return value is a list containing the return values of the particular :py:meth:`submit`
        calls, in an order that corresponds to *job_files*. When an exception was raised during a
        submission, this exception is added to the returned list.
        """
        return self._apply_batch(
            func=self.submit,
            result_type=list,
            job_objs=job_files,
            default_chunk_size=self.chunk_size_submit,
            threads=threads,
            chunk_size=chunk_size,
            callback=callback,
            **kwargs  # noqa
        )

    def cancel_batch(self, job_ids, threads=None, chunk_size=None, callback=None, **kwargs):
        """
        Cancels a batch of jobs given by *job_ids* via a thread pool of size *threads* which
        defaults to its instance attribute. When *chunk_size*, which defaults to
        :py:attr:`chunk_size_cancel`, is not negative, *job_ids* are split into chunks of that size
        which are passed to :py:meth:`cancel`.

        When *callback* is set, it is invoked after each successful job (or job chunk) cancelling
        with the index of the corresponding job id (starting at 0) and either *None* or an exception
        if any occurred. All other *kwargs* are passed to :py:meth:`cancel`.

        Exceptions that occured during job cancelling are stored in a list and returned. An empty
        list means that no exceptions occured.
        """
        results = self._apply_batch(
            func=self.cancel,
            result_type=dict,
            job_objs=job_ids,
            default_chunk_size=self.chunk_size_cancel,
            threads=threads,
            chunk_size=chunk_size,
            callback=callback,
            **kwargs  # noqa
        )

        # return only errors
        return [error for error in results.values() if isinstance(error, Exception)]

    def cleanup_batch(self, job_ids, threads=None, chunk_size=None, callback=None, **kwargs):
        """
        Cleans up a batch of jobs given by *job_ids* via a thread pool of size *threads* which
        defaults to its instance attribute. When *chunk_size*, which defaults to
        :py:attr:`chunk_size_cleanup`, is not negative, *job_ids* are split into chunks of that size
        which are passed to :py:meth:`cleanup`.

        When *callback* is set, it is invoked after each successful job (or job chunk) cleaning with
        the index of the corresponding job id (starting at 0) and either *None* or an exception if
        any occurred. All other *kwargs* are passed to
        :py:meth:`cleanup`.

        Exceptions that occured during job cleaning are stored in a list and returned. An empty list
        means that no exceptions occured.
        """
        results = self._apply_batch(
            func=self.cleanup,
            result_type=dict,
            job_objs=job_ids,
            default_chunk_size=self.chunk_size_cleanup,
            threads=threads,
            chunk_size=chunk_size,
            callback=callback,
            **kwargs  # noqa
        )

        # return only errors
        return [error for error in results.values() if isinstance(error, Exception)]

    def query_batch(self, job_ids, threads=None, chunk_size=None, callback=None, **kwargs):
        """
        Queries the status of a batch of jobs given by *job_ids* via a thread pool of size *threads*
        which defaults to its instance attribute. When *chunk_size*, which defaults to
        :py:attr:`chunk_size_query`, is not negative, *job_ids* are split into chunks of that size
        which are passed to :py:meth:`query`.

        When *callback* is set, it is invoked after each successful job (or job chunk) status query
        with the index of the corresponding job id (starting at 0) and the obtained status query
        data or an exception if any occurred. All other *kwargs* are passed to :py:meth:`query`.

        This method returns a dictionary that maps job ids to either the status query data or to an
        exception if any occurred.
        """
        return self._apply_batch(
            func=self.query,
            result_type=dict,
            job_objs=job_ids,
            default_chunk_size=self.chunk_size_query,
            threads=threads,
            chunk_size=chunk_size,
            callback=callback,
            **kwargs  # noqa
        )

    def _apply_group(
        self,
        func,
        result_type,
        group_func,
        job_objs,
        threads=None,
        callback=None,
        **kwargs  # noqa
    ):
        # default arguments
        threads = max(threads or self.threads or 1, 1)

        # group objects
        job_objs = group_func(make_list(job_objs))

        # factory to call the passed callback for each job file even when chunking
        def cb_factory(i):
            if not callable(callback):
                return None

            def wrapper(result_data):
                offset = sum(map(len, list(job_objs.values())[:i]))
                for j in range(len(list(job_objs.values())[i])):
                    data = result_data if isinstance(result_data, Exception) else result_data[j]
                    callback(offset + j, data)

            return wrapper

        # threaded processing
        pool = ThreadPool(threads)
        results = [
            pool.apply_async(func, make_tuple(arg), kwargs, callback=cb_factory(i))
            for i, arg in enumerate(job_objs.items())
        ]
        pool.close()
        pool.join()

        # store result data or an exception
        result_data = result_type()
        for _job_objs, res in six.moves.zip(job_objs.values(), results):
            data = get_async_result_silent(res)
            for i, job_obj in enumerate(_job_objs):
                if isinstance(result_data, list):
                    result_data.append(data if isinstance(data, Exception) else data[i])
                else:
                    result_data[job_obj] = data if isinstance(data, Exception) else data[job_obj]

        return result_data

    def submit_group(self, job_files, threads=None, callback=None, **kwargs):
        """
        Submits several job groups given by *job_files* via a thread pool of size *threads* which
        defaults to its instance attribute. As per the definition of a job group, a single job file
        can result in multiple jobs being processed on the remote batch system.

        When *callback* is set, it is invoked after each successful job submission with the index of
        the corresponding job (starting at 0) and either the assigned job id or an exception if any
        occurred. All other *kwargs* are passed to :py:meth:`submit`.

        The return value is a list containing the return values of the particular :py:meth:`submit`
        calls, in an order that in general corresponds *job_files*, with ids of single jobs per job
        file properly expanded. When an exception was raised during a submission, this exception is
        added to the returned list.
        """
        # in order to use the generic grouping mechanism in _apply_group create a trivial group_func
        def group_func(job_files):
            groups = defaultdict(list)
            for job_file in job_files:
                groups[job_file].append(job_file)
            return groups

        return self._apply_group(
            func=self.submit,
            result_type=list,
            group_func=group_func,
            job_objs=job_files,
            threads=threads,
            callback=callback,
            **kwargs  # noqa
        )

    def cancel_group(self, job_ids, threads=None, callback=None, **kwargs):
        """
        Takes several *job_ids*, groups them according to :py:meth:`group_job_ids`, and cancels all
        groups simultaneously via a thread pool of size *threads* which defaults to its instance
        attribute.

        When *callback* is set, it is invoked after each successful job cancellation with the index
        of the corresponding job id (starting at 0) and either *None* or an exception if any
        occurred. All other *kwargs* are passed to :py:meth:`cancel`.

        Exceptions that occured during job cancelling are stored in a list and returned. An empty
        list means that no exceptions occured.
        """
        results = self._apply_group(
            func=self.cancel,
            result_type=dict,
            group_func=self.group_job_ids,
            job_objs=job_ids,
            threads=threads,
            callback=callback,
            **kwargs  # noqa
        )

        # return only errors
        return [error for error in results.values() if isinstance(error, Exception)]

    def cleanup_group(self, job_ids, threads=None, callback=None, **kwargs):
        """
        Takes several *job_ids*, groups them according to :py:meth:`group_job_ids`, and cleans up
        all groups simultaneously via a thread pool of size *threads* which defaults to its instance
        attribute.

        When *callback* is set, it is invoked after each successful job cleanup with the index of
        the corresponding job id (starting at 0) and either *None* or an exception if any occurred.
        All other *kwargs* are passed to :py:meth:`cleanup`.

        Exceptions that occured during job cancelling are stored in a list and returned. An empty
        list means that no exceptions occured.
        """
        results = self._apply_group(
            func=self.cleanup,
            result_type=dict,
            group_func=self.group_job_ids,
            job_objs=job_ids,
            threads=threads,
            callback=callback,
            **kwargs  # noqa
        )

        # return only errors
        return [error for error in results.values() if isinstance(error, Exception)]

    def query_group(self, job_ids, threads=None, callback=None, **kwargs):
        """
        Takes several *job_ids*, groups them according to :py:meth:`group_job_ids`, and queries the
        status of all groups simultaneously via a thread pool of size *threads* which defaults to
        its instance attribute.

        When *callback* is set, it is invoked after each successful job status query with the index
        of the corresponding job id (starting at 0) and the obtained status query data or an
        exception if any occurred. All other *kwargs* are passed to :py:meth:`query`.

        This method returns a dictionary that maps job ids to either the status query data or to an
        exception if any occurred.
        """
        return self._apply_group(
            func=self.query,
            result_type=dict,
            group_func=self.group_job_ids,
            job_objs=job_ids,
            threads=threads,
            callback=callback,
            **kwargs  # noqa
        )

    def status_line(self, counts, last_counts=None, sum_counts=None, timestamp=True, align=False,
            color=False):
        """
        Returns a job status line containing job counts per status. When *last_counts* is *True*,
        the status line also contains the differences in job counts with respect to the counts from
        the previous call to this method. When you pass a list or tuple, those values are used
        intead to compute the differences.

        The status line starts with the sum of jobs which is inferred from *counts*. When you want
        to use a custom value, set *sum_counts*. The length of *counts* should match the length of
        *status_names* of this instance. When *timestamp* is *True*, the status line begins with the
        current timestamp. When *timestamp* is a non-empty string, it is used as the ``strftime``
        format.

        *align* handles the alignment of the values in the status line by using a maximum width.
        *True* will result in the default width of 4. When *align* evaluates to *False*, no
        alignment is used. By default, some elements of the status line are colored. Set *color* to
        *False* to disable this feature.

        Example:

        .. code-block:: python

            status_line((2, 0, 0, 0, 0))
            # 12:45:18: all: 2, pending: 2, running: 0, finished: 0, retry: 0, failed: 0

            status_line((0, 2, 0, 0), last_counts=(2, 0, 0, 0), skip=["retry"], timestamp=False)
            # all: 2, pending: 0 (-2), running: 2 (+2), finished: 2 (+0), failed: 0 (+0)
        """
        # check and or set last counts
        use_last_counts = bool(last_counts)
        if use_last_counts and not isinstance(last_counts, (list, tuple)):
            last_counts = self.last_counts or ([0] * len(self.status_names))
        if last_counts and len(last_counts) != len(self.status_names):
            raise Exception("{} last status counts expected, got {}".format(len(self.status_names),
                len(last_counts)))

        # check current counts
        if len(counts) != len(self.status_names):
            raise Exception("{} status counts expected, got {}".format(len(self.status_names),
                len(counts)))

        # calculate differences
        if last_counts:
            diffs = tuple(n - m for n, m in zip(counts, last_counts))

        # number formatting
        if isinstance(align, bool) or not isinstance(align, six.integer_types):
            align = 4 if align else 0
        count_fmt = "%d" if not align else "%{}d".format(align)
        diff_fmt = "%+d" if not align else "%+{}d".format(align)

        # build the status line
        line = ""
        if timestamp:
            time_format = timestamp if isinstance(timestamp, six.string_types) else "%H:%M:%S"
            line += "{}: ".format(time.strftime(time_format))
        if sum_counts is None:
            sum_counts = sum(counts)
        line += "all: " + count_fmt % (sum_counts,)
        for i, (status, count) in enumerate(zip(self.status_names, counts)):
            count = count_fmt % count
            if color:
                count = colored(count, style="bright")
            line += ", {}: {}".format(status, count)

            if diffs:
                diff = diff_fmt % diffs[i]
                if color:
                    # 0 if negative, 1 if zero, 2 if positive
                    style_idx = (diffs[i] > 0) + (diffs[i] >= 0)
                    diff = colored(diff, **self.status_diff_styles[status][style_idx])
                line += " ({})".format(diff)

        # store current counts for next call
        self.last_counts = list(counts)

        return line


class BaseJobFileFactory(six.with_metaclass(ABCMeta, object)):
    """
    Base class that handles the creation of job files. It is likely that inheriting classes only
    need to implement the :py:meth:`create` method as well as extend the constructor to handle
    additional arguments.

    The general idea behind this class is as follows. An instance holds the path to a directory
    *dir*, defaulting to a new, temporary directory inside ``job.job_file_dir`` (which itself
    defaults to the system's tmp path). Job input files, which are supported by almost all job /
    batch systems, are automatically copied into this directory. The file name can be optionally
    postfixed with a configurable string, so that multiple job files can be created and stored
    within the same *dir* without the risk of interfering file names. A common use case would be
    the use of a job number or id. Another *transformation* that is applied to copied files is the
    rendering of variables. For example, when an input file looks like

    .. code-block:: bash

        #!/usr/bin/env bash

        echo "Hello, {{my_variable}}!"

    the rendering mechanism can replace variables such as ``my_variable`` following a double-brace
    notation. Internally, the rendering is implemented in :py:meth:`render_file`, but there is
    usually no need to call this method directly as implementations of this base class might use it
    in their :py:meth:`create` method.

    .. py::classattribute:: config_attrs

        type: list

        List of attributes that is used to create a configuration dictionary. See
        :py:meth:`get_config` for more info.

    .. py::attribute:: dir

        type: string

        The path to the internal job file directory.

    .. py::attribute: cleanup

        type: bool

        Boolean that denotes whether this internal job file directory is temporary and should be
        cleaned up upon instance deletion. It defaults to *True* when the *dir* constructor argument
        is *None*.
    """

    config_attrs = ["dir", "render_variables", "custom_log_file"]

    render_key_cre = re.compile(r"\{\{(\w+)\}\}")

    class Config(object):

        def __repr__(self):
            return repr(self.__dict__)

        def __getattr__(self, attr):
            return self.__dict__[attr]

        def __setattr__(self, attr, value):
            self.__dict__[attr] = value

        def __getitem__(self, attr):
            return self.__dict__[attr]

        def __setitem__(self, attr, value):
            self.__dict__[attr] = value

        def __contains__(self, attr):
            return attr in self.__dict__

    def __init__(self, dir=None, render_variables=None, custom_log_file=None, mkdtemp=None,
            cleanup=None):
        super(BaseJobFileFactory, self).__init__()

        cfg = Config.instance()

        # get default values from config if None
        if mkdtemp is None:
            mkdtemp = cfg.get_expanded_bool("job", "job_file_dir_mkdtemp")
        if cleanup is None:
            cleanup = cfg.get_expanded_bool("job", "job_file_dir_cleanup")

        # store the cleanup flag
        self.cleanup = cleanup

        # when dir ist None, a temporary directory is forced
        if not dir and not mkdtemp:
            mkdtemp = True

        # store the directory, default to the job.job_file_dir config
        self.dir = str(dir or cfg.get_expanded("job", "job_file_dir"))
        self.dir = os.path.expandvars(os.path.expanduser(self.dir))

        # create the directory
        makedirs(self.dir)

        # check if it should be extended by a temporary dir
        if mkdtemp:
            prefix = mkdtemp if isinstance(mkdtemp, six.string_types) else None
            self.dir = tempfile.mkdtemp(dir=self.dir, prefix=prefix)

        # store attributes
        self.render_variables = render_variables or {}
        self.custom_log_file = custom_log_file

        # locks for thread-safe file operations
        self.file_locks = defaultdict(Lock)

    def __del__(self):
        self.cleanup_dir(force=False)

    def __call__(self, *args, **kwargs):
        return self.create(*args, **kwargs)

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        return

    @classmethod
    def postfix_file(cls, path, postfix=None, add_hash=False):
        """
        Adds a *postfix* to a file *path*, right before the first file extension in the base name.
        When *add_hash* is *True*, a hash based on the full source path is added before the postfix.
        Example:

        .. code-block:: python

            postfix_file("/path/to/file.tar.gz", "_1")
            # -> "/path/to/file_1.tar.gz"

            postfix_file("/path/to/file.txt", "_1", add_hash=True)
            # -> "/path/to/file_dacc4374d3_1.txt"

        *postfix* might also be a dictionary that maps patterns to actual postfix strings. When a
        pattern matches the base name of the file, the associated postfix is applied and the path is
        returned. You might want to use an ordered dictionary to control the first match.
        """
        path = str(path)
        dirname, basename = os.path.split(path)

        # get the actual postfix
        _postfix = postfix
        if isinstance(postfix, dict):
            for pattern, _postfix in six.iteritems(postfix):
                if fnmatch.fnmatch(basename, pattern):
                    break
            else:
                _postfix = ""

        # optionally add a hash of the full path
        if add_hash:
            full_path = os.path.realpath(os.path.expandvars(os.path.expanduser(path)))
            _postfix = "_" + create_hash(full_path) + (_postfix or "")

        # add the postfix
        if _postfix:
            parts = basename.split(".", 1)
            parts[0] += _postfix
            path = os.path.join(dirname, ".".join(parts))

        return path

    @classmethod
    def postfix_input_file(cls, path, postfix=None):
        """
        Shorthand for :py:meth:`postfix_file` with *add_hash* set to *True*.
        """
        return cls.postfix_file(path, postfix=postfix, add_hash=True)

    @classmethod
    def postfix_output_file(cls, path, postfix=None):
        """
        Shorthand for :py:meth:`postfix_file` with *add_hash* set to *False*.
        """
        return cls.postfix_file(path, postfix=postfix, add_hash=False)

    @classmethod
    def render_string(cls, s, key, value):
        """
        Renders a string *s* by replacing ``{{key}}`` with *value* and returns it.
        """
        return s.replace("{{" + key + "}}", str(value))

    @classmethod
    def linearize_render_variables(cls, render_variables):
        """
        Linearizes variables contained in the dictionary *render_variables*. In some use cases,
        variables may contain render expressions pointing to other variables, e.g.:

        .. code-block:: python

            render_variables = {
                "variable_a": "Tom",
                "variable_b": "Hello, {{variable_a}}!",
            }

        Situations like this can be simplified by linearizing the variables:

        .. code-block:: python

            linearize_render_variables(render_variables)
            # ->
            # {
            #     "variable_a": "Tom",
            #     "variable_b": "Hello, Tom!",
            # }
        """
        linearized = {}
        for key, value in render_variables.items():
            if not isinstance(value, str):
                raise Exception("render variables must be strings, found '{}' for key '{}'".format(
                    value, key))

            while True:
                m = cls.render_key_cre.search(value)
                if not m:
                    break
                sub_key = m.group(1)
                value = cls.render_string(value, sub_key, render_variables.get(sub_key, ""))
            linearized[key] = value

        # add base64 encoded render variables themselves
        vars_str = base64.b64encode(six.b(json.dumps(linearized) or "-"))
        if six.PY3:
            vars_str = vars_str.decode("utf-8")
        linearized["render_variables"] = vars_str

        return linearized

    @classmethod
    def render_file(cls, src, dst, render_variables, postfix=None, silent=True):
        """
        Renders a source file *src* with *render_variables* and copies it to a new location *dst*.
        In some cases, a render variable value might contain a path that should be subject to file
        postfixing (see :py:meth:`postfix_file`). When *postfix* is not *None*, this method will
        replace substrings in the format ``postfix:<path>`` the postfixed ``path``. In the following
        example, the variable ``my_command`` in *src* will be rendered with a string that contains a
        postfixed path:

        .. code-block:: python

            render_file(src, dst, {"my_command": "echo postfix:some/path.txt"}, postfix="_1")
            # replaces "{{my_command}}" in src with "echo some/path_1.txt" in dst

        In case the file content is not readable, the method returns unless *silent* is *False* in
        which case an exception is raised.
        """
        src, dst = str(src), str(dst)
        if not os.path.isfile(src):
            raise IOError("source file for rendering does not exist: {}".format(src))

        with open(src, "r") as f:
            try:
                content = f.read()
            except UnicodeDecodeError:
                if silent:
                    return
                raise

        def postfix_fn(m):
            return cls.postfix_input_file(m.group(1), postfix=postfix)

        for key, value in six.iteritems(render_variables):
            # value might contain paths to be postfixed, denoted by "__law_job_postfix__:..."
            if postfix:
                value = re.sub(r"\_\_law\_job\_postfix\_\_:([^\s]+)", postfix_fn, value)
            content = cls.render_string(content, key, value)

        # finally, replace all non-rendered keys with empty strings
        content = cls.render_key_cre.sub("", content)

        with open(dst, "w") as f:
            f.write(content)

    @classmethod
    def _expand_template_path(cls, path, variables=None):
        # replace more than three X's with random characters
        if "XXX" in path:
            repl = lambda m: create_random_string(l=len(m.group(1)))
            path = re.sub("(X{3,})", repl, path)

        # replace variables
        if variables:
            for key, value in variables.items():
                path = cls.render_string(path, key, value)

        return path

    def provide_input(self, src, postfix=None, dir=None, render_variables=None,
            skip_existing=False, increment_existing=False):
        """
        Convenience method that copies an input file to a target directory *dir* which defaults to
        the :py:attr:`dir` attribute of this instance. The provided file has the same basename,
        which is optionally postfixed with *postfix*. Essentially, this method calls
        :py:meth:`render_file` when *render_variables* is set, or simply ``shutil.copy2`` otherwise.
        If the file to create is already existing, it is overwritten unless *skip_existing* is
        *True*. If *skip_existing* is *False* but *increment_existing* is *True*, the target path is
        incremented when the file already exists.
        """
        # create the destination path
        src, dir = str(src), dir and str(dir)
        postfixed_src = self.postfix_input_file(src, postfix=postfix)
        dst = os.path.join(os.path.realpath(dir or self.dir), os.path.basename(postfixed_src))

        # check if the file exists but should be skipped
        if skip_existing:
            with self.file_locks[dst]:
                if os.path.exists(dst):
                    return dst

        # check if the path needs to be incremented
        elif increment_existing:
            with self.file_locks[dst]:
                dst = increment_path(dst)

        # provide the file
        with self.file_locks[dst]:
            if render_variables:
                self.render_file(src, dst, render_variables, postfix=postfix)
            else:
                shutil.copy2(src, dst)

        return dst

    def get_config(self, **kwargs):
        """
        The :py:meth:`create` method potentially takes a lot of keywork arguments for configuring
        the content of job files. It is useful if some of these configuration values default to
        attributes that can be set via constructor arguments of this class.

        This method merges keyword arguments *kwargs* (e.g. passed to :py:meth:`create`) with
        default values obtained from instance attributes given in :py:attr:`config_attrs`. It
        returns the merged values in a dictionary that can be accessed via dot-notation (attribute
        notation). Example:

        .. code-block:: python

            class MyJobFileFactory(BaseJobFileFactory):

                config_attrs = ["stdout", "stderr"]

                def __init__(self, stdout="stdout.txt", stderr="stderr.txt", **kwargs):
                    super(MyJobFileFactory, self).__init__(**kwargs)

                    self.stdout = stdout
                    self.stderr = stderr

                def create(self, **kwargs):
                    config = self.get_config(kwargs)

                    # when called as create(stdout="log.txt"):
                    # config.stderr is "stderr.txt"
                    # config.stdout is "log.txt"

                    ...
        """
        cfg = self.Config()
        for attr in self.config_attrs:
            cfg[attr] = copy.deepcopy(kwargs.get(attr, getattr(self, attr)))
        return cfg

    def cleanup_dir(self, force=True):
        """
        Removes the directory that is held by this instance. When *force* is *False*, the directory
        is only removed when :py:attr:`cleanup` is *True*.
        """
        if not self.cleanup and not force:
            return
        if isinstance(self.dir, six.string_types) and os.path.exists(self.dir):
            shutil.rmtree(self.dir)

    @abstractmethod
    def create(self, **kwargs):
        """
        Abstract job file creation method that must be implemented by inheriting classes.
        """
        return


class JobArguments(object):
    """
    Wrapper class for job arguments. Currently, it stores a task class *task_cls*, a list of
    *task_params*, a list of covered *branches*, an *auto_retry* flag, and custom *dashboard_data*.
    It also handles argument encoding as reqired by the job wrapper script at
    `law/job/job.sh <https://github.com/riga/law/blob/master/law/job/job.sh>`__.

    .. py:attribute:: task_cls

        type: :py:class:`law.Register`

        The task class.

    .. py:attribute:: task_params

        type: list

        The list of task parameters.

    .. py:attribute:: branches

       type: list

        The list of branch numbers covered by the task.

    .. py:attribute:: workers

        type: int

        The number of workers to use in "law run" commands.

    .. py:attribute:: auto_retry

        type: bool

        A flag denoting if the job-internal automatic retry mechanism should be used.

    .. py:attribute:: dashboard_data

        type: list

        If a job dashboard is used, this is a list of configuration values as returned by
        :py:meth:`law.job.dashboard.BaseJobDashboard.remote_hook_data`.
    """

    def __init__(self, task_cls, task_params, branches, workers=1, auto_retry=False,
            dashboard_data=None):
        super(JobArguments, self).__init__()

        self.task_cls = task_cls
        self.task_params = task_params
        self.branches = branches
        self.workers = max(workers, 1)
        self.auto_retry = auto_retry
        self.dashboard_data = dashboard_data or []

    @classmethod
    def encode_bool(cls, b):
        """
        Encodes a boolean *b* into a string (``"yes"`` or ``"no"``).
        """
        return "yes" if b else "no"

    @classmethod
    def encode_string(cls, s):
        """
        Encodes a string *s* via base64 encoding.
        """
        encoded = base64.b64encode(six.b(s or "-"))
        return encoded.decode("utf-8") if six.PY3 else encoded

    @classmethod
    def encode_list(cls, l):
        """
        Encodes a list *l* into a string via base64 encoding.
        """
        encoded = base64.b64encode(six.b(" ".join(str(v) for v in l) or "-"))
        return encoded.decode("utf-8") if six.PY3 else encoded

    def get_args(self):
        """
        Returns the list of encoded job arguments. The order of this list corresponds to the
        arguments expected by the job wrapper script.
        """
        return [
            self.task_cls.__module__,
            self.task_cls.__name__,
            self.encode_string(self.task_params),
            self.encode_list(self.branches),
            self.workers,
            self.encode_bool(self.auto_retry),
            self.encode_list(self.dashboard_data),
        ]

    def join(self):
        """
        Returns the list of job arguments from :py:meth:`get_args`, joined into a single string
        using a single space character.
        """
        return " ".join(str(arg) for arg in self.get_args())


class JobInputFile(object):
    """
    Wrapper around a *path* referring to an input file of a job, accompanied by optional flags that
    control how the file should be handled during job submission (mostly within
    :py:meth:`BaseJobFileFactory.provide_input`). See the attributs below for more info.

    .. py:attribute:: path

        type: str

        The path of the input file.

    .. py:attribute:: copy

        type: bool

        Whether this file should be copied into the job submission directory or not.

    .. py:attribute:: share

        type: bool

        Whether the file can be shared in the job submission directory. A shared file is copied only
        once into the submission directory and :py:attr:`render_local` must be *False*.

    .. py:attribute:: forward

        type: bool

        Whether this file should actually not be listed as a normal input file in job description
        but just passed to the list of inputs for treatment in the law job script itself. Only
        considered if supported by the submission system (e.g. local ones such as htcondor or
        slurm).

    .. py:attribute:: increment

        type: bool

        Whether the file path should be incremented when copied if a file with the same name already
        exists in the same submission directory.

    .. py:attribute:: postfix

        type: bool

        Whether the file path should be postfixed when copied.

    .. py:attribute:: render_local

        type: bool

        Whether render variables should be resolved locally when copied.

    .. py:attribute:: render_job

        type: bool

        Whether render variables should be resolved as part of the job script.

    .. py:attribute:: is_remote

        type: bool (read-only)

        Whether the path has a non-empty protocol referring to a remote resource.

    .. py:attribute:: path_sub_abs

        type: str, None

        Absolute file path as seen by the submission node. Set only during job file creation.

    .. py:attribute:: path_sub_rel

        type: str, None

        File path relative to the submission directory if the submission itself is not forced to use
        absolute paths. Otherwise identical to :py:attr:`path_sub_abs`. Set only during job file
        creation.

    .. py:attribute:: path_job_pre_render

        type: str, None

        File path as seen by the job node, prior to a potential job-side rendering. It is a full,
        absolute path in case forwarding is supported, and a relative basename otherwise. Set only
        during job file creation.

    .. py:attribute:: path_job_post_render

        type: str, None

        File path as seen by the job node, after a potential job-side rendering. Therefore, it is
        identical to :py:attr:`path_job_pre_render` if rendering is disabled, and a relative
        basename otherwise. Set only during job file creation.
    """

    def __init__(self, path, copy=None, share=None, forward=None, increment=None, postfix=None,
            render=None, render_local=None, render_job=None):
        super(JobInputFile, self).__init__()

        # when path is a job file instance itself, use its values instead
        if isinstance(path, JobInputFile):
            copy = path.copy
            share = path.share
            forward = path.forward
            increment = path.increment
            postfix = path.postfix
            render_local = path.render_local
            render_job = path.render_job
            path = path.path

        # path must not be a remote file target
        if isinstance(path, RemoteTarget):
            raise ValueError("{}.path should not point to a remote target: {}".format(
                self.__class__.__name__, path))

        # convenience
        if render is not None and render_local is None and render_job is None:
            render_local = bool(render)
            render_job = False

        # set some attributes if undefined, based on most common use cases
        maybe_set = lambda current, default: default if current is None else current
        if copy is not None and not copy:
            # share = maybe_set(share, False)
            share = maybe_set(share, False)
            postfix = maybe_set(postfix, False)
            render_local = maybe_set(render_local, False)
        if share:
            copy = maybe_set(copy, True)
            forward = maybe_set(forward, False)
            render_local = maybe_set(render_local, False)
            postfix = maybe_set(postfix, False)
        if forward:
            copy = maybe_set(copy, False)
            share = maybe_set(share, False)
            render_local = maybe_set(render_local, False)
            postfix = maybe_set(postfix, False)
        if postfix:
            copy = maybe_set(copy, True)
            share = maybe_set(share, False)
            forward = maybe_set(forward, False)
        if render_local:
            copy = maybe_set(copy, True)
            share = maybe_set(share, False)
            forward = maybe_set(forward, False)
            render_job = maybe_set(render_job, False)
        if render_job:
            forward = maybe_set(forward, False)
            render_local = maybe_set(render_local, False)

        # store attributes, apply residual defaults
        self.path = os.path.abspath(os.path.expandvars(os.path.expanduser(get_path(path))))
        self.copy = True if copy is None else bool(copy)
        self.share = False if share is None else bool(share)
        self.forward = False if forward is None else bool(forward)
        self.increment = False if increment is None else bool(increment)
        self.postfix = True if postfix is None else bool(postfix)
        self.render_local = True if render_local is None else bool(render_local)
        self.render_job = False if render_job is None else bool(render_job)

        # some residual attribute checks
        if not self.copy and self.postfix:
            logger.warning(
                "input file at {} is configured not to be copied into the submission directory, "
                "but postfixing is enabled which has no effect".format(self.path),
            )
        if not self.copy and self.share:
            logger.warning(
                "input file at {} is configured not to be copied into the submission directory, "
                "but sharing is enabled which has no effect".format(self.path),
            )
        if self.copy and self.forward:
            logger.warning(
                "input file at {} is configured to be copied into the submission directory, but "
                "but forwarding is enabled which has no effect".format(self.path),
            )
        if not self.copy and self.render_local:
            logger.warning(
                "input file at {} is configured not to be copied into the submission directory, "
                "but rendering is enabled which has no effect".format(self.path),
            )
        if self.share and self.render_local:
            logger.error(
                "input file at {} is configured to be shared across jobs but local rendering is "
                "active, potentially resulting in wrong file content".format(self.path),
            )
        if self.render_local and self.render_job:
            logger.warning(
                "input file at {} is configured to be rendered locally and within the job, which "
                "is likely unnecessary".format(self.path),
            )

        # different path variants as seen by jobs
        self.path_sub_abs = None
        self.path_sub_rel = None
        self.path_job_pre_render = None
        self.path_job_post_render = None

    def __str__(self):
        return self.path

    def __repr__(self):
        return "<{}({}) at {}>".format(
            self.__class__.__name__,
            ", ".join("{}={}".format(attr, getattr(self, attr)) for attr in [
                "path", "copy", "share", "forward", "increment", "postfix", "render_local",
                "render_job",
            ]),
            hex(id(self)),
        )

    def __eq__(self, other):
        # check equality via path comparison
        if isinstance(other, JobInputFile):
            return self.path == other.path
        return self.path == str(other)

    @property
    def is_remote(self):
        return get_scheme(self.path) not in ("file", None)


class DeprecatedInputFiles(dict):
    """
    Class to keep track of input files for remote jobs that is only used to show a deprecation
    warning for users still relying on lists. Therefore, this class emulates the most used list
    methods and internally fills input files using dict methods. To be removed in version 1.0.
    """

    @classmethod
    def _log_warning(cls, method):
        logger.warning_once(
            "the use of input_files.{} is deprecated, please consider updating your code towards "
            "using dictionaries instead, e.g., 'input_files[key] = path'; by doing so, law "
            "automatically adds a render variable 'key' that will refer to the postfixed path of "
            "the input file for immediate use in remote jobs".format(method),
        )

    def __init__(self, *args, **kwargs):
        paths = None
        if not kwargs and len(args) == 1 and isinstance(args[0], list):
            paths = args[0]
            args = ()

        super(DeprecatedInputFiles, self).__init__(*args, **kwargs)

        if paths:
            self.extend(paths)

    def _append(self, path):
        # generate a key by taking the basename of the path and strip the file extension
        path = str(path)
        key = os.path.basename(path).split(".", 1)[0]
        while key in self:
            key += "_"

        self[key] = path

    def append(self, path):
        # deprecation warning until v0.1
        self._log_warning("append(path)")
        self._append(path)

    def extend(self, paths):
        # deprecation warning until v0.1
        self._log_warning("extend([path, ...])")
        for path in paths:
            self._append(path)

    def __add__(self, paths):
        # type-preserving shallow copy
        self_ = self.__class__(self)
        self_.extend(paths)
        return self_

    def __iadd__(self, paths):
        self.extend(paths)
        return self



# coding: utf-8

"""
Cache for remote files on local disk.
"""

__all__ = ["RemoteCache"]


import os
import shutil
import time
import tempfile
import weakref
import atexit
from contextlib import contextmanager

from law.config import Config
from law.util import (
    makedirs, human_bytes, parse_bytes, parse_duration, create_hash, user_owns_file, io_lock,
)
from law.logger import get_logger


logger = get_logger(__name__)


class RemoteCache(object):

    TMP = "__TMP__"

    lock_postfix = ".lock"

    _instances = []

    def __new__(cls, *args, **kwargs):
        inst = object.__new__(cls)

        # cache instances
        cls._instances.append(inst)

        return inst

    @classmethod
    def cleanup_all(cls):
        # clear all caches
        for inst in cls._instances:
            try:
                inst._cleanup()
            except:
                pass

    @classmethod
    def parse_config(cls, section, config=None, overwrite=False):
        from law.sandbox.base import _sandbox_switched

        # reads a law config section and returns parsed file system configs
        cfg = Config.instance()

        if config is None:
            config = {}

        # helper to add a config value if it exists, extracted with a config parser method
        def add(option, func):
            cache_option = "cache_" + option
            if cfg.is_missing_or_none(section, cache_option):
                return
            elif option not in config or overwrite:
                config[option] = func(section, cache_option)

        def get_size(section, cache_option):
            value = cfg.get_expanded(section, cache_option)
            return parse_bytes(value, input_unit="MB", unit="MB")

        def get_time(section, cache_option):
            value = cfg.get_expanded(section, cache_option)
            return parse_duration(value, input_unit="s", unit="s")

        add("root", cfg.get_expanded)
        add("cleanup", cfg.get_expanded_bool)
        add("max_size", get_size)
        add("mtime_patience", cfg.get_expanded_float)
        add("file_perm", cfg.get_expanded_int)
        add("dir_perm", cfg.get_expanded_int)
        add("wait_delay", get_time)
        add("max_waits", cfg.get_expanded_int)
        add("global_lock", cfg.get_expanded_bool)

        # inside sandboxes, never cleanup since the outer process will do that if needed
        if _sandbox_switched:
            config["cleanup"] = False

        return config

    def __init__(self, fs, root=TMP, cleanup=False, max_size=0, mtime_patience=1.0,
            file_perm=0o0660, dir_perm=0o0770, wait_delay=5.0, max_waits=120, global_lock=False):
        object.__init__(self)
        # max_size is in MB, wait_delay is in seconds

        # create a unique name based on fs attributes
        name = "{}_{}".format(fs.__class__.__name__, create_hash(fs.base[0]))

        # create the root dir, handle tmp
        root = os.path.expandvars(os.path.expanduser(str(root))) or self.TMP
        if not os.path.exists(root) and root == self.TMP:
            cfg = Config.instance()
            tmp_dir = cfg.get_expanded("target", "tmp_dir")
            base = tempfile.mkdtemp(dir=tmp_dir)
            cleanup = True
        else:
            base = os.path.join(root, name)
            makedirs(base, dir_perm)

        # save attributes and configs
        self.root = root
        self.fs_ref = weakref.ref(fs)
        self.base = base
        self.name = name
        self.cleanup = cleanup
        self.max_size = max_size
        self.mtime_patience = mtime_patience
        self.dir_perm = dir_perm
        self.file_perm = file_perm
        self.wait_delay = wait_delay
        self.max_waits = max_waits
        self.global_lock = global_lock

        # path to the global lock file which should guard global actions such as cache allocations
        self._global_lock_path = self._lock_path(os.path.join(base, "global"))

        # currently locked cache paths, only used to clean up broken files during cleanup
        self._locked_cpaths = set()

        logger.debug("created {} at '{}'".format(self.__class__.__name__, self.base))

    def __del__(self):
        try:
            self._cleanup()
        except (OSError, TypeError):
            pass

    def __repr__(self):
        return "<{} '{}' at {}>".format(self.__class__.__name__, self.base, hex(id(self)))

    def __contains__(self, rpath):
        return os.path.exists(self.cache_path(rpath))

    @property
    def fs(self):
        return self.fs_ref()

    def _cleanup(self):
        # full cleanup or remove open locks
        if getattr(self, "cleanup", False):
            if os.path.exists(self.base):
                shutil.rmtree(self.base)
        else:
            for cpath in set(self._locked_cpaths):
                self._unlock(cpath)
                self._remove(cpath)
            self._locked_cpaths.clear()
            self._unlock_global()
        logger.debug("cleanup RemoteCache at '{}'".format(self.base))

    def cache_path(self, rpath):
        rpath = str(rpath)
        basename = "{}_{}".format(create_hash(rpath), os.path.basename(rpath))
        return os.path.join(self.base, basename)

    def _lock_path(self, cpath):
        return str(cpath) + self.lock_postfix

    def is_locked_global(self):
        return os.path.exists(self._global_lock_path)

    def _is_locked(self, cpath):
        return os.path.exists(self._lock_path(cpath))

    def is_locked(self, rpath):
        return self._is_locked(self.cache_path(rpath))

    def _unlock_global(self):
        try:
            os.remove(self._global_lock_path)
        except OSError:
            pass

    def _unlock(self, cpath):
        try:
            os.remove(self._lock_path(cpath))
        except OSError:
            pass

    def _await_global(self, delay=None, max_waits=None, silent=False):
        delay = delay if delay is not None else self.wait_delay
        max_waits = max_waits if max_waits is not None else self.max_waits
        _max_waits = max_waits

        while self.is_locked_global():
            if max_waits <= 0:
                if not silent:
                    raise Exception("max_waits of {} exceeded while waiting for global lock".format(
                        _max_waits))
                return False

            time.sleep(delay)
            max_waits -= 1

        return True

    def _await(self, cpath, delay=None, max_waits=None, silent=False, global_lock=None):
        cpath = str(cpath)
        delay = delay if delay is not None else self.wait_delay
        max_waits = max_waits if max_waits is not None else self.max_waits
        _max_waits = max_waits
        global_lock = self.global_lock if global_lock is None else global_lock

        # strategy: wait as long the file is locked and if the file size did not change, reduce
        # max_waits per iteration and raise when 0 is reached
        last_size = -1
        while self._is_locked(cpath) or (global_lock and self.is_locked_global()):
            if max_waits <= 0:
                if not silent:
                    raise Exception("max_waits of {} exceeded while waiting for file '{}'".format(
                        _max_waits, cpath))
                return False

            time.sleep(delay)

            # only reduce max_waits when the file size did not change
            # otherwise, set it to its original value again
            if os.path.exists(cpath):
                size = os.stat(cpath).st_size
                if size != last_size:
                    last_size = size
                    max_waits = _max_waits + 1

            max_waits -= 1

        return True

    @contextmanager
    def _lock_global(self, **kwargs):
        self._await_global(**kwargs)

        try:
            with io_lock:
                with open(self._global_lock_path, "w") as f:
                    f.write("")
                os.utime(self._global_lock_path, None)

            yield
        finally:
            self._unlock_global()

    @contextmanager
    def _lock(self, cpath, **kwargs):
        cpath = str(cpath)
        lock_path = self._lock_path(cpath)

        self._await(cpath, **kwargs)

        try:
            with io_lock:
                with open(lock_path, "w") as f:
                    f.write("")
                self._locked_cpaths.add(cpath)
                try:
                    os.utime(lock_path, None)
                except OSError:
                    pass

            yield
        except:
            # when something went really wrong, conservatively delete the cached file
            self._remove(cpath, lock=False)
            raise
        finally:
            # unlock again
            self._unlock(cpath)
            if cpath in self._locked_cpaths:
                self._locked_cpaths.remove(cpath)

    def lock(self, rpath):
        return self._lock(self.cache_path(rpath))

    def allocate(self, size):
        logger.debug("allocating {0[0]:.2f} {0[1]} in cache '{1}'".format(human_bytes(size), self))

        # determine stats and current cache size
        file_stats = []
        for elem in os.listdir(self.base):
            if elem.endswith(self.lock_postfix):
                continue
            cpath = os.path.join(self.base, elem)
            file_stats.append((cpath, os.stat(cpath)))
        current_size = sum(stat.st_size for _, stat in file_stats)

        # get the available space of the disk that contains the cache in bytes, leave 10%
        fs_stat = os.statvfs(self.base)
        free_size = fs_stat.f_frsize * fs_stat.f_bavail * 0.9

        # determine the maximum size of the cache
        # make sure it is always smaller than what is available
        if self.max_size <= 0:
            max_size = current_size + free_size
        else:
            max_size = min(self.max_size * 1024**2, current_size + free_size)

        # determine the size of files that need to be deleted
        delete_size = current_size + size - max_size
        if delete_size <= 0:
            logger.debug("cache space sufficient, {0[0]:.2f} {0[1]} remaining".format(
                human_bytes(-delete_size)))
            return True

        logger.info("need to delete {0[0]:.2f} {0[1]} from cache".format(
            human_bytes(delete_size)))

        # delete files, ordered by their access time, skip locked ones
        for cpath, cstat in sorted(file_stats, key=lambda tpl: tpl[1].st_atime):
            if self._is_locked(cpath):
                continue
            self._remove(cpath)
            delete_size -= cstat.st_size
            if delete_size <= 0:
                return True

        logger.warning("could not allocate remaining {0[0]:.2f} {0[1]} in cache".format(
            human_bytes(delete_size)))

        return False

    def _touch(self, cpath, times=None):
        cpath = str(cpath)
        if os.path.exists(cpath):
            if user_owns_file(cpath):
                os.chmod(cpath, self.file_perm)
            os.utime(cpath, times)

    def touch(self, rpath, times=None):
        self._touch(self.cache_path(rpath), times=times)

    def _mtime(self, cpath):
        return os.stat(str(cpath)).st_mtime

    def mtime(self, rpath):
        return self._mtime(self.cache_path(rpath))

    def check_mtime(self, rpath, rmtime):
        if self.mtime_patience < 0:
            return True

        return abs(self.mtime(rpath) - rmtime) <= self.mtime_patience

    def _remove(self, cpath, lock=True):
        def remove():
            try:
                os.remove(str(cpath))
            except OSError:
                pass

        if lock:
            with self._lock(cpath):
                remove()
        else:
            remove()

    def remove(self, rpath, lock=True):
        return self._remove(self.cache_path(rpath), lock=lock)


atexit.register(RemoteCache.cleanup_all)



# coding: utf-8

"""
Dropbox file system and targets based on the GFAL file interface.
"""

__all__ = ["DropboxFileSystem", "DropboxTarget", "DropboxFileTarget", "DropboxDirectoryTarget"]


import six

import law
from law.config import Config
from law.target.remote import (
    RemoteFileSystem, RemoteTarget, RemoteFileTarget, RemoteDirectoryTarget,
)
from law.logger import get_logger


logger = get_logger(__name__)

law.contrib.load("gfal")


class DropboxFileSystem(RemoteFileSystem):

    file_interface_cls = law.gfal.GFALFileInterface

    @classmethod
    def parse_config(cls, section, config=None, overwrite=False):
        config = super(DropboxFileSystem, cls).parse_config(section, config=config,
            overwrite=overwrite)

        cfg = Config.instance()

        # helper to add a config value if it exists, extracted with a config parser method
        def add(option, func, prefix="dropbox_"):
            if option not in config or overwrite:
                config[option] = func(section, prefix + option)

        # app key
        add("app_key", cfg.get_expanded)

        # app secret
        add("app_secret", cfg.get_expanded)

        # access token
        add("access_token", cfg.get_expanded)

        return config

    def __init__(self, section=None, app_key=None, app_secret=None, access_token=None, **kwargs):
        # read configs from section and combine them with kwargs to get the file system and
        # file interface configs
        section, fs_config, fi_config = self._init_configs(section, "default_dropbox_fs",
            "_dropbox_fs_defaults", kwargs)

        # store the config section
        self.config_section = section

        # overwrite dropbox credentials with passed values
        if app_key:
            fs_config["app_key"] = app_key
        if app_secret:
            fs_config["app_secret"] = app_secret
        if access_token:
            fs_config["access_token"] = access_token

        # base path, app key, app secret and access token are mandatory
        msg = ("attribute '{{0}}' must not be empty, set it either directly in the {} constructor, "
            "or add the option '{{0}}' to your config section '{}'".format(
                self.__class__.__name__, self.config_section))
        if not fi_config.get("base"):
            raise Exception(msg.format("base"))
        for attr in ["app_key", "app_secret", "access_token"]:
            if not fs_config.get(attr):
                raise Exception(msg.format(attr))

        # enforce some configs
        fs_config["has_permissions"] = False

        # pass dropbox credentials via gfal options
        gfal_options = fi_config.setdefault("gfal_options", {})
        gfal_options.setdefault("integer", []).append(("DROPBOX", "OAUTH", 2))
        gfal_options.setdefault("string", []).extend([
            ("DROPBOX", "APP_KEY", str(fs_config.pop("app_key"))),
            ("DROPBOX", "APP_SECRET", str(fs_config.pop("app_secret"))),
            ("DROPBOX", "ACCESS_TOKEN", str(fs_config.pop("access_token"))),
        ])

        # create the file interface
        file_interface = self.file_interface_cls(**fi_config)

        # initialize the file system itself
        super(DropboxFileSystem, self).__init__(file_interface, **fs_config)


# try to set the default fs instance
try:
    DropboxFileSystem.default_instance = DropboxFileSystem()
    logger.debug("created default DropboxFileSystem instance '{}'".format(
        DropboxFileSystem.default_instance))
except Exception as e:
    logger.warning("could not create default DropboxFileSystem instance: {}".format(e))


class DropboxTarget(RemoteTarget):
    """ __init__(path, fs=DropboxFileSystem.default_instance, **kwargs)
    """

    def __init__(self, path, fs=DropboxFileSystem.default_instance, **kwargs):
        if fs is None:
            fs = DropboxFileSystem.default_instance
        elif isinstance(fs, six.string_types):
            fs = DropboxFileSystem(fs)
        RemoteTarget.__init__(self, path, fs, **kwargs)


class DropboxFileTarget(DropboxTarget, RemoteFileTarget):

    pass


class DropboxDirectoryTarget(DropboxTarget, RemoteDirectoryTarget):

    pass


DropboxTarget.file_class = DropboxFileTarget
DropboxTarget.directory_class = DropboxDirectoryTarget
